# pbp, Apache License 2.0
# Filename: metadata/generator/gen_iclisten.py
# Description:  Captures ICListen wav metadata in a pandas dataframe from either a local directory or S3 bucket.

import re
from datetime import timedelta
from datetime import datetime
from typing import List

import boto3

import pandas as pd
from pathlib import Path
from progressbar import progressbar
import pbp.json_generator.utils as utils
from pbp.json_generator.corrector import MetadataCorrector
from pbp.json_generator.metadata_extractor import GenericWavFile
from pbp.json_generator.gen_abstract import MetadataGeneratorAbstract


class IcListenMetadataGenerator(MetadataGeneratorAbstract):
    log_prefix = None

    def __init__(
        self,
        log,  # : loguru.Logger,
        uri: str,
        json_base_dir: str,
        start: datetime,
        end: datetime,
        prefix: List[str],
        seconds_per_file: float = 600.0,
    ):
        """
        Captures ICListen wav metadata in a pandas dataframe from either a local directory or S3 bucket.
        :param uri:
            The local directory or S3 bucket that contains the wav files
        :param json_base_dir:
            The local directory to store the metadata
        :param start:
            The start date to search for wav files
        :param end:
            The end date to search for wav files
        :param prefix:
            The search pattern to match the wav files, e.g. 'MARS' for MARS_YYYYMMDD_HHMMSS.wav
        :param seconds_per_file:
            The number of seconds per file expected in a wav file to check for missing data. If 0, then no check is done.
        :return:
        """
        super().__init__(log, uri, json_base_dir, prefix, start, end, seconds_per_file)
        self.log_prefix = f"{self.__class__.__name__} {start:%Y%m%d}"

    def run(self):
        self.log.info(
            f"{self.log_prefix} Generating metadata for {self.start} to {self.end}..."
        )

        bucket_name, prefix, scheme = utils.parse_s3_or_gcp_url(self.audio_loc)

        # gs is not supported for icListen
        if scheme == "gs":
            self.log.error(
                f"{self.log_prefix} GS is not supported for icListen audio files"
            )
            return

        # Run for each day in the range
        for day in pd.date_range(self.start, self.end, freq="D"):
            try:
                self.df = None
                self.log.info(
                    f"{self.log_prefix} Searching in {self.audio_loc}/*.wav for wav files that match the search pattern {self.prefix}* ..."
                )

                wav_files = []

                def check_file(f: str, f_start_dt: datetime, f_end_dt: datetime):
                    """
                    Check if the file matches the search pattern and is within the start and end dates
                    :param f:
                        The path to the file
                    :param f_start_dt:
                        The start date to check
                    :param f_end_dt:
                        The end date to check
                    :return:
                    """

                    f_path = Path(f)
                    f_wav_dt = None

                    for s in self.prefix:
                        # see if the file is a regexp match to search
                        rc = re.search(s, f_path.stem)

                        if rc and rc.group(0):
                            try:
                                # MARS file date is in the filename MARS_YYYYMMDD_HHMMSS.wav
                                f_path_dt = datetime.strptime(
                                    f_path.stem, f"{s}%Y%m%d_%H%M%S"
                                )

                                if f_start_dt <= f_path_dt <= f_end_dt:
                                    self.log.info(
                                        f"{self.log_prefix} Found {f_path.name} to process"
                                    )
                                    wav_files.append(
                                        GenericWavFile(self.log, f, f_path_dt)
                                    )
                                    f_wav_dt = f_path_dt
                            except ValueError:
                                self.log.error(
                                    f"{self.log_prefix} Could not parse {f_path.name}"
                                )
                                return None

                    return f_wav_dt

                # Set the start and end dates to 1 hour before and after the start and end dates
                start_dt = day - timedelta(hours=1)
                end_dt = day + timedelta(days=1)

                if scheme == "file":
                    wav_path = Path(self.audio_loc)
                    for filename in progressbar(
                        sorted(wav_path.rglob("*.wav")), prefix="Searching : "
                    ):
                        check_file(filename.as_posix(), start_dt, end_dt)
                if scheme == "s3":
                    client = boto3.client("s3")
                    for day_hour in pd.date_range(start=start_dt, end=end_dt, freq="h"):
                        bucket = f"{bucket_name}-{day_hour.year:04d}"
                        prefix = f"{day_hour.month:02d}/MARS_{day_hour.year:04d}{day_hour.month:02d}{day_hour.day:02d}_{day_hour.hour:02d}"
                        paginator = client.get_paginator("list_objects")

                        operation_parameters = {"Bucket": bucket, "Prefix": prefix}
                        page_iterator = paginator.paginate(**operation_parameters)
                        self.log.info(
                            f"{self.log_prefix}  Searching in bucket: {bucket} prefix: {prefix}"
                        )
                        # list the objects in the bucket
                        # loop through the objects and check if they match the search pattern
                        for page in page_iterator:
                            if "Contents" not in page:
                                self.log.info(
                                    f"{self.log_prefix}  No data found in {bucket}"
                                )
                                break

                            for obj in page["Contents"]:
                                key = obj["Key"]
                                wav_dt = check_file(
                                    f"s3://{bucket}/{key}", start_dt, end_dt
                                )
                                if wav_dt is None:
                                    continue
                                if wav_dt > end_dt or wav_dt < start_dt:
                                    break

                self.log.info(
                    f"{self.log_prefix}  Found {len(wav_files)} files to process that cover the period {start_dt} - {end_dt}"
                )

                # sort the files by start time
                wav_files.sort(key=lambda x: x.start)

                # create a dataframe from the wav files
                self.log.info(
                    f"{self.log_prefix}  Creating dataframe from {len(wav_files)} files spanning {wav_files[0].start} to {wav_files[-1].start}..."
                )
                for wc in wav_files:
                    df_wav = wc.to_df()

                    # concatenate the metadata to the dataframe
                    self.df = pd.concat([self.df, df_wav], axis=0)

                self.log.debug(f"{self.log_prefix}  Running metadata corrector for {day}")
                corrector = MetadataCorrector(
                    self.log,
                    self.df,
                    self.json_base_dir,
                    day,
                    utils.InstrumentType.NRS,
                    True,
                    self.seconds_per_file,
                )
                corrector.run()

            except Exception as ex:
                self.log.exception(str(ex))


if __name__ == "__main__":
    from pbp.logging_helper import create_logger

    log_dir = Path("tests/log")
    json_dir = Path("tests/json/mars")
    log_dir.mkdir(exist_ok=True, parents=True)
    json_dir.mkdir(exist_ok=True, parents=True)

    log = create_logger(
        log_filename_and_level=(
            f"{log_dir}/test_iclisten_metadata_generator.log",
            "INFO",
        ),
        console_level="INFO",
    )

    start = datetime(2023, 7, 18, 0, 0, 0)
    end = datetime(2023, 7, 18, 0, 0, 0)

    # If only running one day, use a single generator
    generator = IcListenMetadataGenerator(
        log,
        uri="s3://pacific-sound-256khz",
        json_base_dir=json_dir.as_posix(),
        prefix=["MARS"],
        start=start,
        end=end,
        seconds_per_file=600,
    )
    generator.run()



from pbp.main_args import parse_arguments

# Some imports, in particular involving data processing, cause a delay that is
# noticeable when just running the --help option. We get around this issue by
# postponing the imports until actually needed. See the main() function.


def main():
    opts = parse_arguments()

    # pylint: disable=import-outside-toplevel
    import os

    from pbp.file_helper import FileHelper
    from pbp.logging_helper import create_logger
    from pbp.process_helper import ProcessHelper

    log = create_logger(
        log_filename_and_level=(
            f"{opts.output_dir}/{opts.output_prefix}{opts.date}.log",
            "INFO",
        ),
        console_level="WARNING",
    )

    s3_client = None
    if opts.s3:
        # pylint: disable=import-outside-toplevel
        import boto3

        kwargs = {}
        aws_region = os.getenv("AWS_REGION")
        if aws_region is not None:
            kwargs["region_name"] = aws_region

        s3_client = boto3.client("s3", **kwargs)

    gs_client = None
    if opts.gs:
        # pylint: disable=import-outside-toplevel
        from google.cloud.storage import Client as GsClient

        # TODO credentials; for now assuming only anonymous downloads
        gs_client = GsClient.create_anonymous_client()

    file_helper = FileHelper(
        log=log,
        json_base_dir=opts.json_base_dir,
        audio_base_dir=opts.audio_base_dir,
        audio_path_map_prefix=opts.audio_path_map_prefix,
        audio_path_prefix=opts.audio_path_prefix,
        s3_client=s3_client,
        gs_client=gs_client,
        download_dir=opts.download_dir,
        assume_downloaded_files=opts.assume_downloaded_files,
        retain_downloaded_files=opts.retain_downloaded_files,
    )

    process_helper = ProcessHelper(
        log=log,
        file_helper=file_helper,
        output_dir=opts.output_dir,
        output_prefix=opts.output_prefix,
        global_attrs_uri=opts.global_attrs,
        set_global_attrs=opts.set_global_attrs,
        variable_attrs_uri=opts.variable_attrs,
        voltage_multiplier=opts.voltage_multiplier,
        sensitivity_uri=opts.sensitivity_uri,
        sensitivity_flat_value=opts.sensitivity_flat_value,
        max_segments=opts.max_segments,
        subset_to=tuple(opts.subset_to) if opts.subset_to else None,
    )
    try:
        process_helper.process_day(opts.date)
    except KeyboardInterrupt:
        log.info("INTERRUPTED")


if __name__ == "__main__":
    main()



#!/usr/bin/env python

#
# TODO Adjustments for GS as this script is still only focused on S3.
#

# Script for cloud based processing. By this, we basically mean the ability
# to get input files (json and wav) from S3 and write output files to S3.
#
# Inputs to the program are to be passed via environment variables:
#  DATE: (Required)
#     The date to process. Format: "YYYYMMDD".
#  S3_JSON_BUCKET_PREFIX: (Optional)
#     Bucket prefix to be used to locate the YYYYMMDD.json file
#     By default, "s3://pacific-sound-metadata/256khz".
#  S3_OUTPUT_BUCKET: (Optional)
#     The bucket to write the generated output to.
#     Typically this is to be provided but it is optional to facilitate testing.
#  OUTPUT_PREFIX: (Optional)
#     Output filename prefix. By default, "milli_psd_".
#     The resulting file will be named as <OUTPUT_PREFIX><DATE>.nc.
#  GLOBAL_ATTRS_URI: (Optional)
#     URI of JSON file with global attributes to be added to the NetCDF file.
#  VARIABLE_ATTRS_URI: (Optional)
#     URI of JSON file with attributes to associate to the variables in the NetCDF file.
#  VOLTAGE_MULTIPLIER: (Optional)
#     Applied on the loaded signal.
#  SENSITIVITY_NETCDF_URI: (Optional)
#     URI of sensitivity NetCDF file that should be used to calibrate the result.
#  SENSITIVITY_FLAT_VALUE: (Optional)
#     Flat sensitivity value to be used for calibration
#     if SENSITIVITY_NETCDF_URI is not given.
#  SUBSET_TO: (Required)  Format: "lower,upper"
#     Subset the resulting PSD to [lower, upper), in terms of central frequency.
#
# *Note*:
#   TODO retrieve sensitivity information using PyHydrophone when none
#     of the `SENSITIVITY_*` environment variables above are given.
#
# Mainly for testing purposes, also these environment variables are inspected:
#
#  CLOUD_TMP_DIR: (Optional)
#     Local workspace for downloads and for generated files to be uploaded.
#     By default, "cloud_tmp".
#
#  MAX_SEGMENTS: (Optional)
#     0, the default, means no restriction, that is, all segments for each day
#     will be processed.
#
#  ASSUME_DOWNLOADED_FILES: (Optional)
#     If "yes", then if any destination file for a download exists,
#     it is assumed downloaded already.
#     The default is that downloads are always performed.
#
#  RETAIN_DOWNLOADED_FILES: (Optional)
#     If "yes", do not remove any downloaded files after use.
#     The default is that any downloaded file is removed after use.

import os
import pathlib

import boto3

from pbp.file_helper import FileHelper
from pbp.logging_helper import create_logger
from pbp.process_helper import ProcessHelper


def main():
    # --------------------------
    # Cloud preparations:

    # The date to process. Format: "YYYYMMDD"
    date = os.environ["DATE"]

    # Bucket prefix to be used to locate the YYYYMMDD.json file
    json_bucket_prefix = os.getenv(
        "S3_JSON_BUCKET_PREFIX", "s3://pacific-sound-metadata/256khz"
    )

    output_prefix = os.getenv("OUTPUT_PREFIX", "milli_psd_")

    # The bucket to write the output to
    output_bucket = os.getenv("S3_OUTPUT_BUCKET")

    # Applied on the loaded signal
    voltage_multiplier = (
        float(os.getenv("VOLTAGE_MULTIPLIER"))
        if os.getenv("VOLTAGE_MULTIPLIER") is not None
        else None
    )

    global_attrs_uri = os.getenv("GLOBAL_ATTRS_URI")
    variable_attrs_uri = os.getenv("VARIABLE_ATTRS_URI")

    # URI of sensitivity NetCDF file to be used for calibration
    sensitivity_uri = os.getenv("SENSITIVITY_NETCDF_URI")

    # Flat sensitivity value to be used for calibration
    sensitivity_flat_value = (
        float(os.getenv("SENSITIVITY_FLAT_VALUE"))
        if os.getenv("SENSITIVITY_FLAT_VALUE") is not None
        else None
    )

    subset_to_string = os.environ["SUBSET_TO"]
    subset_to = tuple(int(val.strip()) for val in subset_to_string.split(","))
    assert len(subset_to) == 2

    # Convenience for testing (0 means no restriction)
    max_segments = int(os.getenv("MAX_SEGMENTS", "0"))

    # workspace for downloads and generated files to be uploaded
    cloud_tmp_dir = os.getenv("CLOUD_TMP_DIR", "cloud_tmp")

    download_dir = f"{cloud_tmp_dir}/downloads"
    pathlib.Path(download_dir).mkdir(parents=True, exist_ok=True)

    generated_dir = f"{cloud_tmp_dir}/generated"
    pathlib.Path(generated_dir).mkdir(parents=True, exist_ok=True)

    log_filename = f"{generated_dir}/{output_prefix}{date}.log"
    log = create_logger(
        log_filename_and_level=(log_filename, "INFO"),
        console_level="DEBUG",
    )

    kwargs = {}
    aws_region = os.getenv("AWS_REGION")
    if aws_region is not None:
        kwargs["region_name"] = aws_region

    s3_client = boto3.client("s3", **kwargs)

    if output_bucket is not None:
        # create output_bucket if it does not exist
        found = any(
            b["Name"] == output_bucket for b in s3_client.list_buckets()["Buckets"]
        )
        if not found:
            log.info(f"Creating bucket {output_bucket}")
            s3_client.create_bucket(
                Bucket=output_bucket,
                CreateBucketConfiguration={"LocationConstraint": aws_region},
            )

    else:
        log.info("No output bucket specified. Output will not be uploaded.")

    # --------------------------
    # Get working:

    file_helper = FileHelper(
        json_base_dir=json_bucket_prefix,
        s3_client=s3_client,
        gs_client=None,  # TODO
        download_dir=download_dir,
        assume_downloaded_files=os.getenv("ASSUME_DOWNLOADED_FILES", "no") == "yes",
        retain_downloaded_files=os.getenv("RETAIN_DOWNLOADED_FILES", "no") == "yes",
    )

    process_helper = ProcessHelper(
        file_helper=file_helper,
        output_dir=generated_dir,
        output_prefix=output_prefix,
        global_attrs_uri=global_attrs_uri,
        variable_attrs_uri=variable_attrs_uri,
        voltage_multiplier=voltage_multiplier,
        sensitivity_uri=sensitivity_uri,
        sensitivity_flat_value=sensitivity_flat_value,
        max_segments=max_segments,
        subset_to=subset_to,
    )

    result = process_helper.process_day(date)

    if result is None:
        log.warning(f"No NetDF file was generated.  ({date=})")
        return

    log.info(f"Generated files: {result.generated_filenames}")

    if output_bucket is not None:

        def upload(filename):
            log.info(f"Uploading {filename} to {output_bucket}")
            filename_out = pathlib.Path(filename).name
            ok = s3_client.upload_file(filename, output_bucket, filename_out)
            log.info(f"Upload result: {ok}")

        for generated_filename in result.generated_filenames:
            upload(generated_filename)

        # result of uploading the log itself won't of course show up there
        upload(log_filename)

    else:
        log.info("No uploads attempted as output bucket was not given.")


if __name__ == "__main__":
    main()



# pbp, Apache License 2.0
# Filename: json_generator/metadata_extractor.py
# Description: Utilities for wav file metadata reading. Supports SoundTrap, NRS and and icListen audio files

from pathlib import Path
from typing import Optional

from six.moves.urllib.request import urlopen
import io
import re
import soundfile as sf
import pandas as pd
from datetime import datetime, timedelta
import xml.etree.ElementTree as ET
from pbp.json_generator.utils import parse_s3_or_gcp_url


class AudioFile:
    def __init__(self, path_or_url: str, start: datetime):
        """
        Abstract class for reading wav file metadata
        :param path_or_url:
            The path or url to the wav file
        :param start:
            The start time of the wav file
        """
        self.start = start
        self.path_or_url = path_or_url
        self.end: Optional[datetime] = None
        self.duration_secs = 0.0
        self.fs = -1
        self.frames = -1
        self.channels = -1
        self.subtype = ""
        self.exception = ""

    def has_exception(self):
        return True if len(self.exception) > 0 else False

    def to_df(self):
        # if the self.path_or_url is a url, then add to the data frame with the appropriate prefix
        if "s3://" in self.path_or_url or "gs://" in self.path_or_url:
            df = pd.DataFrame(
                {
                    "uri": self.path_or_url,
                    "start": self.start,
                    "end": self.end,
                    "fs": self.fs,
                    "duration_secs": self.duration_secs,
                    "channels": self.channels,
                    "subtype": self.subtype,
                    "exception": self.exception,
                },
                index=[self.start],
            )
        else:
            df = pd.DataFrame(
                {
                    "uri": "file://" + self.path_or_url,
                    "start": self.start,
                    "end": self.end,
                    "fs": self.fs,
                    "duration_secs": self.duration_secs,
                    "channels": self.channels,
                    "subtype": self.subtype,
                    "exception": self.exception,
                },
                index=[self.start],
            )
        return df

    def get_max_freq(self):
        return self.fs / 2


class SoundTrapWavFile(AudioFile):
    def __init__(self, path_or_url: str, xml_file: str, start: datetime):
        """
        SoundTrapWavFile uses the metadata from the xml files, not the wav file itself
        :param path_or_url:
            The path or uri of the wav file
        :param xml_file:
            The uri of the xml file that contains the metadata
        :param path_or_url:

        :param start:
        """
        super().__init__(path_or_url, start)
        tree = ET.parse(xml_file)
        root = tree.getroot()
        wav_start_dt = None
        wav_stop_dt = None
        sample_count = None

        # Iterate over the XML elements grabbing the needed metadata values
        for element in root.iter("WavFileHandler"):
            # Get the value of the id attribute
            value = element.get("SamplingStartTimeUTC")
            if value:
                wav_start_dt = datetime.strptime(value, "%Y-%m-%dT%H:%M:%S")

            value = element.get("SamplingStopTimeUTC")
            if value:
                wav_stop_dt = datetime.strptime(value, "%Y-%m-%dT%H:%M:%S")

            value = element.get("SampleCount")
            if value:
                sample_count = int(value)

        # Error checking
        if not wav_start_dt or not wav_stop_dt or not sample_count:
            raise ValueError(f"Error reading {xml_file}. Missing metadata")

        self.path_or_url = path_or_url
        self.start = wav_start_dt
        self.end = wav_stop_dt
        self.duration_secs = sample_count / 48000
        self.fs = 48000
        self.frames = sample_count
        self.channels = 1
        self.subtype = "SoundTrap"
        self.exception = ""  # no exceptions for SoundTrap  files


class GenericWavFile(AudioFile):
    """GenericWavFile uses the metadata from the wav file itself,
    but only grabs the needed metadata from the header in S3"""

    def __init__(self, log, path_or_url: str, start: datetime):
        super().__init__(path_or_url, start)
        self.log = log
        self.path_or_url = path_or_url
        self.start = start
        self.duration_secs = 0
        self.fs = -1
        self.frames = -1
        self.channels = -1
        self.subtype = ""
        self.exception = ""
        self.path_or_url = path_or_url
        bytes_per_sec = (
            3 * 256e3
        )  # 3 bytes per sample at 24-bit resolution and 256 kHz sampling rate

        try:
            # if the in_file is a s3 url, then read the metadata from the s3 url
            if re.match(r"^s3://", path_or_url):
                p = Path(path_or_url)
                bucket, key = p.parts[1], "/".join(p.parts[2:])
                url = f"http://{bucket}.s3.amazonaws.com/{key}"

                # read the first 20,000 bytes of the file to get the metadata
                info = sf.info(io.BytesIO(urlopen(url).read(20_000)), verbose=True)
                # get the duration from the extra_info data field which stores the duration in total bytes
                fields = info.extra_info.split()
                idx = fields.index("data")
                self.duration_secs = float(fields[idx + 2]) / bytes_per_sec
                # get the size in bytes of the data+RIFF header
                idx = fields.index("RIFF")
                riff_size = int(fields[idx + 2]) + 8
                # get the content length from the http header
                content_length = int(urlopen(url).info()["Content-Length"])
                # if the content length is less than the size of the data+RIFF header, then the file is truncated but
                # still may be usable
                if content_length < riff_size:
                    self.exception = f"Truncated file {path_or_url}. Content length {content_length} < RIFF size {riff_size}"
                    # calculate the duration which is the size of the content length minus the size of the RIFF
                    # header which is 44 bytes. Round the duration to the nearest second since the recording is
                    # always in 1 second increments
                    self.duration_secs = round(content_length - 44) / bytes_per_sec
                    self.log.warning(self.exception)
            else:
                info = sf.info(path_or_url)
                self.duration_secs = info.duration

            self.end = self.start + timedelta(microseconds=int(self.duration_secs * 1e6))
            self.fs = info.samplerate
            self.frames = info.frames
            self.channels = info.channels
            self.subtype = info.subtype if info.subtype else ""
        except Exception as ex:
            self.log.exception(f"Corrupt file {path_or_url}. {ex}")


class FlacFile(AudioFile):
    """FlacFile uses the metadata from the flac file itself,
    but only grabs the needed metadata from the header in gs or local file system."""

    def __init__(self, log, path_or_url: str, start: datetime):
        super().__init__(path_or_url, start)
        self.log = log
        self.path_or_url = path_or_url
        self.start = start
        self.end: Optional[datetime] = None
        self.duration_secs = 0
        self.fs = -1
        self.frames = -1
        self.channels = -1
        self.subtype = ""
        self.exception = ""
        self.path_or_url = path_or_url

        try:
            # if the in_file is a gs url, then read the metadata
            bucket, prefix, scheme = parse_s3_or_gcp_url(path_or_url)
            if scheme == "gs":
                url = f"http://storage.googleapis.com/{bucket}/{prefix}"

                info = sf.info(io.BytesIO(urlopen(url).read(20_000)), verbose=True)

                # get the duration from the extra_info data field which stores the duration in total bytes
                fields = info.extra_info.split(":")
                sample_rate = int(fields[3].split("\n")[0])
                channels = int(fields[2].split("\n")[0])
                length_microseconds = int(info.frames * 1e6 / info.samplerate)
                # get the file name from the url
                file_name = url.split("/")[-1]

                # files are in the format NRS11_20191231_230836.flac'
                # extract the timestamp from the file name
                f = Path(file_name).stem.split("_")
                # If the last two digits of the timestamp are 60, subtract 1 second
                # This is a bug in the FlacFile name
                if f[2][-2:] == "60":
                    f_c = f[1] + f[2]
                    # Make the last two digits 59
                    f_c = f_c[:-2] + "59"
                else:
                    f_c = f[1] + f[2]
                # convert the timestamp to a datetime object
                timestamp = datetime.strptime(f_c, "%Y%m%d%H%M%S")
                self.start = timestamp
                self.end = self.start + timedelta(microseconds=length_microseconds)
                self.duration_secs = int(length_microseconds / 1e6)
                self.channels = channels
                self.subtype = "flac"
                self.fs = sample_rate
                self.frames = info.frames if info.frames else 0
            if scheme == "file" or scheme == "":
                info = sf.info(path_or_url)
                length_microseconds = int(info.frames * 1e6 / info.samplerate)
                self.duration_secs = int(length_microseconds / 1e6)
                self.end = self.start + timedelta(microseconds=length_microseconds)
                self.fs = info.samplerate
                self.frames = info.frames
                self.channels = info.channels
                self.subtype = info.subtype if info.subtype else ""
        except Exception as ex:
            self.log.exception(f"Corrupt file {path_or_url}. {ex}")



def get_pbp_version() -> str:
    try:
        import importlib.metadata

        return importlib.metadata.version("mbari-pbp")
    except Exception:  # pylint: disable=broad-exception-caught
        try:
            from poetry.core.factory import Factory

            factory = Factory()
            poetry = factory.create_poetry()
            return poetry.package.version.text
        except Exception:  # pylint: disable=broad-exception-caught
            return "version not found"


def get_pypam_version() -> str:
    try:
        import importlib.metadata

        return importlib.metadata.version("lifewatch-pypam")
    except Exception:  # pylint: disable=broad-exception-caught
        return "??"



from argparse import ArgumentParser, RawTextHelpFormatter

from pbp import get_pbp_version
from pbp.json_generator.utils import InstrumentType


def parse_arguments():
    description = (
        "Generate JSONs with audio metadata for NRS flac files, "
        "IcListen wav files, and Soundtrap wav files from either a local directory or gs/s3 bucket."
    )
    example = """
Examples:
    pbp-json-gen \\
                 --json-base-dir=tests/json/nrs \\
                 --output-dir=output \\
                 --uri=s3://pacific-sound-ch01 \\
                 --start=20220902 \\
                 --end=20220902 \\
                 --prefix=MARS \\
                 --recorder=NRS
    """

    parser = ArgumentParser(
        description=description, epilog=example, formatter_class=RawTextHelpFormatter
    )

    parser.add_argument(
        "--version",
        action="version",
        version=get_pbp_version(),
    )

    parser.add_argument(
        "--recorder",
        choices=[InstrumentType.NRS, InstrumentType.ICLISTEN, InstrumentType.SOUNDTRAP],
        required=True,
        help="Choose the audio instrument type",
    )

    parser.add_argument(
        "--json-base-dir",
        type=str,
        metavar="dir",
        required=True,
        help="JSON base directory to store the metadata",
    )

    parser.add_argument(
        "--output-dir",
        type=str,
        metavar="dir",
        required=True,
        help="Output directory to store logs",
    )

    parser.add_argument(
        "--uri",
        type=str,
        metavar="uri",
        required=True,
        default=None,
        help="Location of the audio files. S3 location supported for IcListen or Soundtrap, and GS supported for NRS.",
    )

    parser.add_argument(
        "--start",
        type=str,
        required=True,
        metavar="YYYYMMDD",
        help="The starting date to be processed.",
    )

    parser.add_argument(
        "--end",
        type=str,
        required=True,
        metavar="YYYYMMDD",
        help="The ending date to be processed.",
    )

    parser.add_argument(
        "--prefix",
        type=str,
        required=True,
        nargs="+",
        help="Prefix for search to match the audio files. Assumption is the prefix is separated by an "
        "underscore, e.g. 'MARS_'.",
    )

    return parser.parse_args()



# pbp, Apache License 2.0
# Filename: json_generator/utils.py
# Description:  Utility functions for parsing S3, GS or local file urls and defining sound instrument types for metadata generation
from typing import Tuple
from urllib.parse import urlparse


class InstrumentType:
    NRS = "NRS"
    ICLISTEN = "ICLISTEN"
    SOUNDTRAP = "SOUNDTRAP"


def parse_s3_or_gcp_url(url) -> Tuple[str, str, str]:
    """
    Parse the S3, GS of local file url
    :param url:
    :return:
    """
    parsed_url = urlparse(url)
    bucket = parsed_url.netloc
    prefix = parsed_url.path.lstrip("/")
    if parsed_url.scheme == "file":
        bucket = ""
        prefix = parsed_url.path
    return bucket, prefix, parsed_url.scheme



import pathlib
from dataclasses import dataclass
from datetime import datetime, timezone

from typing import Any, List, Optional, OrderedDict, Tuple

import numpy as np
import xarray as xr

from pbp import get_pbp_version, get_pypam_version
from pbp.file_helper import FileHelper
from pbp.metadata import MetadataHelper, parse_attributes, replace_snippets
from pbp.misc_helper import gen_hour_minute_times, parse_date
from pbp.pypam_support import ProcessResult, PypamSupport


@dataclass
class ProcessDayResult:
    """
    The result returned from `process_day`.

    Contains the list of paths to generated files
    (NetCDF and others depending on given parameters)
    as well as the generated dataset.
    """

    generated_filenames: list[str]
    dataset: xr.Dataset


class ProcessHelper:
    def __init__(
        self,
        log,  # : loguru.Logger,
        file_helper: FileHelper,
        output_dir: str,
        output_prefix: str,
        gen_netcdf: bool = True,
        global_attrs_uri: Optional[str] = None,
        set_global_attrs: Optional[list[list[str]]] = None,
        variable_attrs_uri: Optional[str] = None,
        voltage_multiplier: Optional[float] = None,
        sensitivity_uri: Optional[str] = None,
        sensitivity_flat_value: Optional[float] = None,
        max_segments: int = 0,
        subset_to: Optional[Tuple[int, int]] = None,
    ):
        """

        :param file_helper:
            File loader.
        :param output_dir:
            Output directory.
        :param output_prefix:
            Output filename prefix.
        :param gen_netcdf:
            True to generate the netCDF file.
        :param global_attrs_uri:
            URI of JSON file with global attributes to be added to the NetCDF file.
        :param set_global_attrs:
            List of [key, value] pairs to be considered for the global attributes.
        :param variable_attrs_uri:
            URI of JSON file with variable attributes to be added to the NetCDF file.
        :param voltage_multiplier:
            Applied on the loaded signal.
        :param sensitivity_uri:
            URI of sensitivity NetCDF for calibration of result.
            Has precedence over `sensitivity_flat_value`.
        :param sensitivity_flat_value:
            Flat sensitivity value to be used for calibration.
        :param max_segments:
            Maximum number of segments to process for each day.
            By default, 0 (no limit).
        :param subset_to:
            Tuple of (lower, upper) frequency limits to use for the PSD,
            lower inclusive, upper exclusive.
        """
        self.log = log

        self.log.info(
            "Creating ProcessHelper:"
            + f"\n    output_dir:             {output_dir}"
            + f"\n    output_prefix:          {output_prefix}"
            + f"\n    gen_netcdf:             {gen_netcdf}"
            + f"\n    global_attrs_uri:       {global_attrs_uri}"
            + f"\n    set_global_attrs:       {set_global_attrs}"
            + f"\n    variable_attrs_uri:     {variable_attrs_uri}"
            + f"\n    voltage_multiplier:     {voltage_multiplier}"
            + f"\n    sensitivity_uri:        {sensitivity_uri}"
            + f"\n    sensitivity_flat_value: {sensitivity_flat_value}"
            + (
                f"\n    max_segments:           {max_segments}"
                if max_segments > 0
                else ""
            )
            + f"\n    subset_to:              {subset_to}"
            + "\n"
        )
        self.file_helper = file_helper
        self.output_dir = output_dir
        self.output_prefix = output_prefix
        self.gen_netcdf = gen_netcdf

        self.metadata_helper = MetadataHelper(
            self.log,
            self._load_attributes("global", global_attrs_uri, set_global_attrs),
            self._load_attributes("variable", variable_attrs_uri),
        )

        self.max_segments = max_segments
        self.subset_to = subset_to

        self.voltage_multiplier: Optional[float] = voltage_multiplier

        self.sensitivity_da: Optional[xr.DataArray] = None
        self.sensitivity_flat_value: Optional[float] = sensitivity_flat_value

        if sensitivity_uri is not None:
            s_local_filename = file_helper.get_local_filename(sensitivity_uri)
            if s_local_filename is not None:
                sensitivity_ds = xr.open_dataset(s_local_filename)
                self.log.info(f"Will use loaded sensitivity from {s_local_filename=}")
                self.sensitivity_da = sensitivity_ds.sensitivity
                self.log.debug(f"{self.sensitivity_da=}")
            else:
                self.log.error(
                    f"Unable to resolve sensitivity_uri: '{sensitivity_uri}'. Ignoring it."
                )

        if self.sensitivity_da is None and self.sensitivity_flat_value is not None:
            self.log.info(
                f"Will use given flat sensitivity value: {sensitivity_flat_value}"
            )

        self.pypam_support = PypamSupport(self.log)

        pathlib.Path(output_dir).mkdir(exist_ok=True)

    def _load_attributes(
        self,
        what: str,
        attrs_uri: Optional[str],
        set_attrs: Optional[list[list[str]]] = None,
    ) -> Optional[OrderedDict[str, Any]]:
        if attrs_uri:
            self.log.info(f"Loading {what} attributes from {attrs_uri=}")
            filename = self.file_helper.get_local_filename(attrs_uri)
            if filename is not None:
                with open(filename, "r", encoding="UTF-8") as f:
                    res = parse_attributes(f.read(), pathlib.Path(filename).suffix)
                    for k, v in set_attrs or []:
                        res[k] = v
                    return res
            else:
                self.log.error(f"Unable to resolve '{attrs_uri=}'. Ignoring it.")
        else:
            self.log.info(f"No '{what}' attributes URI given.")
        return None

    def process_day(self, date: str) -> Optional[ProcessDayResult]:
        """
        Generates NetCDF file with the result of processing all segments of the given day.

        :param date:
            Date to process in YYYYMMDD format.
        :return:
            ProcessDayResult, or None if no segments at all were processed for the day.
        """
        year, month, day = parse_date(date)
        if not self.file_helper.select_day(year, month, day):
            return None

        at_hour_and_minutes: List[Tuple[int, int]] = list(
            gen_hour_minute_times(self.file_helper.segment_size_in_mins)
        )

        if self.max_segments > 0:
            at_hour_and_minutes = at_hour_and_minutes[: self.max_segments]
            self.log.info(f"NOTE: Limiting to {len(at_hour_and_minutes)} segments ...")

        self.process_hours_minutes(at_hour_and_minutes)

        result: Optional[ProcessResult] = self.pypam_support.process_captured_segments(
            sensitivity_da=self.sensitivity_da,
        )

        if result is None:
            self.log.warning(
                f"No segments processed, nothing to aggregate for day {date}."
            )
            return None

        psd_da = result.psd_da

        # rename 'frequency_bins' dimension to 'frequency':
        psd_da = psd_da.swap_dims(frequency_bins="frequency")

        data_vars = {
            "psd": psd_da,
            "effort": result.effort_da,
        }

        if self.sensitivity_da is not None:
            freq_subset = self.sensitivity_da.interp(frequency=psd_da.frequency)
            data_vars["sensitivity"] = freq_subset

        elif self.sensitivity_flat_value is not None:
            # better way to capture a scalar?
            data_vars["sensitivity"] = xr.DataArray(
                data=[self.sensitivity_flat_value],
                dims=["1"],
            ).astype(np.float32)

        md_helper = self.metadata_helper

        md_helper.add_variable_attributes(psd_da["time"], "time")
        md_helper.add_variable_attributes(data_vars["effort"], "effort")
        md_helper.add_variable_attributes(psd_da["frequency"], "frequency")
        if "sensitivity" in data_vars:
            md_helper.add_variable_attributes(data_vars["sensitivity"], "sensitivity")
        md_helper.add_variable_attributes(data_vars["psd"], "psd")

        ds_result = xr.Dataset(
            data_vars=data_vars,
            attrs=self._get_global_attributes(year, month, day),
        )

        generated_filenames = []
        basename = f"{self.output_dir}/{self.output_prefix}{year:04}{month:02}{day:02}"

        if self.gen_netcdf:
            nc_filename = f"{basename}.nc"
            save_dataset_to_netcdf(self.log, ds_result, nc_filename)
            generated_filenames.append(nc_filename)

        self.file_helper.day_completed()

        return ProcessDayResult(generated_filenames, ds_result)

    def process_hours_minutes(self, hour_and_minutes: List[Tuple[int, int]]):
        self.log.info(f"Processing {len(hour_and_minutes)} segments ...")
        for at_hour, at_minute in hour_and_minutes:
            self.process_segment_at_hour_minute(at_hour, at_minute)

    def process_segment_at_hour_minute(self, at_hour: int, at_minute: int):
        file_helper = self.file_helper
        year, month, day = file_helper.year, file_helper.month, file_helper.day
        assert year is not None and month is not None and day is not None

        dt = datetime(year, month, day, at_hour, at_minute, tzinfo=timezone.utc)

        self.log.debug(
            f"Segment at {at_hour:02}h:{at_minute:02}m ...\n"
            + f"  - extracting {file_helper.segment_size_in_mins * 60}-sec segment:"
        )
        extraction = file_helper.extract_audio_segment(at_hour, at_minute)
        if extraction is None:
            self.log.warning(f"cannot get audio segment at {at_hour:02}:{at_minute:02}")
            self.pypam_support.add_missing_segment(dt)
            return

        audio_info, audio_segment = extraction

        if self.pypam_support.parameters_set:
            if self.pypam_support.fs != audio_info.samplerate:
                self.log.info(
                    f"ERROR: samplerate changed from {self.pypam_support.fs} to {audio_info.samplerate}"
                )
                return
        else:
            self.log.info("Got audio parameters")
            self.pypam_support.set_parameters(
                audio_info.samplerate,
                subset_to=self.subset_to,
            )

        if self.voltage_multiplier is not None:
            audio_segment *= self.voltage_multiplier

        if self.sensitivity_flat_value is not None:
            # convert signal to uPa
            audio_segment = audio_segment * 10 ** (self.sensitivity_flat_value / 20)

        self.pypam_support.add_segment(dt, audio_segment)

    def _get_global_attributes(self, year: int, month: int, day: int):
        coverage_date = f"{year:04}-{month:02}-{day:02}"
        global_attrs = {
            "time_coverage_start": f"{coverage_date} 00:00:00Z",
            "time_coverage_end": f"{coverage_date} 23:59:00Z",
            "date_created": datetime.utcnow().strftime("%Y-%m-%d"),
        }
        md_helper = self.metadata_helper
        md_helper.set_some_global_attributes(global_attrs)
        snippets = {
            "{{PBP_version}}": get_pbp_version(),
            "{{PyPAM_version}}": get_pypam_version(),
        }
        global_attrs = md_helper.get_global_attributes()
        # for each, key, have the {{key}} snippet for replacement
        # in case it is used in any values:
        for k, v in global_attrs.items():
            snippets["{{" + k + "}}"] = v
        return replace_snippets(global_attrs, snippets)


def save_dataset_to_netcdf(
    log,  #: loguru.Logger,
    ds: xr.Dataset,
    filename: str,
) -> bool:
    log.info(f"  - saving dataset to: {filename}")
    try:
        ds.to_netcdf(
            filename,
            engine="h5netcdf",
            encoding={
                "effort": {"_FillValue": None},
                "frequency": {"_FillValue": None},
                "sensitivity": {"_FillValue": None},
            },
        )
        return True
    except Exception as e:  # pylint: disable=broad-exception-caught
        error = f"Unable to save {filename}: {e}"
        log.error(error)
        print(error)
        return False



from datetime import datetime
from pathlib import Path

from pbp.json_generator.gen_nrs import NRSMetadataGenerator
from pbp.json_generator.gen_iclisten import IcListenMetadataGenerator
from pbp.json_generator.gen_soundtrap import SoundTrapMetadataGenerator
from pbp.main_json_generator_args import parse_arguments

# Some imports, in particular involving data processing, cause a delay that is
# noticeable when just running the --help option. We get around this issue by
# postponing the imports until actually needed. See the main() function.


def main():
    opts = parse_arguments()

    # pylint: disable=import-outside-toplevel
    from pbp.logging_helper import create_logger

    log = create_logger(
        log_filename_and_level=(
            f"{opts.output_dir}/{opts.recorder}{opts.start}_{opts.end}.log",
            "INFO",
        ),
        console_level="INFO",
    )

    log_dir = Path(opts.output_dir)
    json_dir = Path(opts.json_base_dir)
    log_dir.mkdir(exist_ok=True, parents=True)
    json_dir.mkdir(exist_ok=True, parents=True)
    start = datetime.strptime(opts.start, "%Y%m%d")
    end = datetime.strptime(opts.end, "%Y%m%d")

    try:
        if opts.recorder == "NRS":
            generator = NRSMetadataGenerator(
                log=log,
                uri=opts.uri,
                json_base_dir=json_dir.as_posix(),
                prefix=opts.prefix,
                start=start,
                end=end,
            )
            generator.run()
        if opts.recorder == "ICLISTEN":
            generator = IcListenMetadataGenerator(
                log=log,
                uri=opts.uri,
                json_base_dir=json_dir.as_posix(),
                prefix=opts.prefix,
                start=start,
                end=end,
            )
            generator.run()
            # TODO: add multiprocessing here for speed-up
        if opts.recorder == "SOUNDTRAP":
            generator = SoundTrapMetadataGenerator(
                log=log,
                uri=opts.uri,
                json_base_dir=json_dir.as_posix(),
                prefix=opts.prefix,
                start=start,
                end=end,
            )
            generator.run()
    except KeyboardInterrupt:
        log.info("INTERRUPTED")


if __name__ == "__main__":
    main()



from typing import Final

DEFAULT_LAT_LON_FOR_SOLPOS: Final = (36.7128, -122.186)

DEFAULT_TITLE: Final = (
    "MBARI MARS Cabled Observatory   Monterey Bay, California, USA"
    f"   {DEFAULT_LAT_LON_FOR_SOLPOS[0]} °N, {DEFAULT_LAT_LON_FOR_SOLPOS[1]} °W"
)

DEFAULT_YLIM: Final = (10, 100000)

DEFAULT_CMLIM: Final = (32, 108)

DEFAULT_DPI: Final = 200



# pbp, Apache License 2.0
# Filename: metadata/utils/corrector.py
# Description: Correct metadata for wav files and saves the results to a json file.

import datetime
from datetime import timedelta
import numpy as np
import pandas as pd
from pathlib import Path
import shutil
import tempfile
import json

from pbp.json_generator.utils import InstrumentType


class MetadataCorrector:
    def __init__(
        self,
        log,  # : loguru.Logger,
        raw_df: pd.DataFrame,
        json_path_out: str,
        day: datetime.datetime,
        instrument_type: InstrumentType,
        time_correct: bool = False,
        seconds_per_file: float = -1,
    ):
        """
        Correct the metadata for a day and save to a json file
        :param raw_df:
            The dataframe containing the raw metadata to correct
        :param json_path_out:
            The path to save the corrected metadata json file
        :param day:
            The day to correct
        :param instrument_type:
            The type of instrument the metadata is coming from: NRS, ICLISTEN, SOUNDTRAP
        :param time_correct:
            True if need to adjust the time stamp based only supported for ICLISTEN
        :param seconds_per_file:
            (optional) number of seconds in each file
        """
        self.instrument_type = instrument_type
        self.raw_df = raw_df
        self.json_base_dir = json_path_out
        self.day = day
        self.log = log
        self.time_correct = time_correct
        self.seconds_per_file = seconds_per_file
        self.files_per_day = None
        # Must have seconds per file for ICLISTEN to correct for drift conditional check
        if self.instrument_type == InstrumentType.ICLISTEN:
            if self.seconds_per_file == -1:
                self.log.exception("No seconds per file provided for ICLISTEN")
                return
            self.files_per_day = int(86400 / self.seconds_per_file)
            self.log.debug(
                f"Metadata corrector for {self.instrument_type} with {self.seconds_per_file} seconds per file"
            )

    def run(self):
        """Run the corrector"""

        try:
            # Filter the metadata to the day capturing the files both immediately
            # before and after the day
            next_day = self.day + timedelta(days=1)
            day_df = self.raw_df[
                ((self.raw_df["start"] >= self.day) & (self.raw_df["start"] <= next_day))
                | ((self.raw_df["end"] >= self.day) & (self.raw_df["start"] < self.day))
            ]

            self.log.debug(
                f"Creating metadata for day {self.day} from {len(day_df)} files..."
            )

            if len(day_df) == 0:
                self.log.warning(f"No metadata found for day {self.day}")
                return

            # convert the start and end times to datetime
            day_df = day_df.copy()
            day_df["start"] = pd.to_datetime(day_df["start"])
            day_df["end"] = pd.to_datetime(day_df["end"])

            # get the file list that covers the requested day
            self.log.info(
                f'Found {len(day_df)} files from day {self.day}, starting {day_df.iloc[0]["start"]} ending {day_df.iloc[-1]["end"]}'
            )

            # if there are no files, then return
            if len(day_df) == 0:
                self.log.warning(f"No files found for {self.day}")
                return

            for index, row in day_df.iterrows():
                self.log.debug(f'File {row["uri"]} duration {row["duration_secs"]} ')
                if (
                    self.seconds_per_file > 0
                    and row["duration_secs"] != self.seconds_per_file
                ):
                    self.log.warning(
                        f'File {row["duration_secs"]}  != {self.seconds_per_file}. File is not complete'
                    )

            # check whether there is a discrepancy between the number of seconds in the file and the number
            # of seconds in the metadata. If there is a discrepancy, then correct the metadata
            # This is only reliable for full days of data contained in complete files for IcListen data
            day_df["jitter_secs"] = 0

            if self.instrument_type == InstrumentType.ICLISTEN and (
                len(day_df) == self.files_per_day + 1
                and len(day_df["duration_secs"].unique()) == 1
                and day_df.iloc[0]["duration_secs"] == self.seconds_per_file
            ):
                # check whether the differences are all the same
                if len(day_df["start"].diff().unique()) == 1 or self.time_correct:
                    self.log.warning(f"No drift for {self.day}")
                else:
                    self.log.info(f"Correcting drift for {self.day}")

                    # correct the metadata
                    jitter = 0
                    start = day_df.iloc[0]["start"]
                    end = start + timedelta(seconds=self.seconds_per_file)

                    for index, row in day_df.iterrows():
                        # jitter is the difference between the expected start time and the actual start time
                        # jitter is 0 for the first file
                        if row.start == start:
                            # round the jitter to the nearest second
                            jitter = start.to_datetime64() - row.start.to_datetime64()
                            jitter = int(jitter / np.timedelta64(1, "s"))

                        # correct the start and end times
                        day_df.loc[index, "start"] = start
                        day_df.loc[index, "end"] = end
                        day_df.loc[index, "jitter_secs"] = jitter

                        if self.time_correct:
                            end = row.end
                        else:
                            end = start + timedelta(seconds=self.seconds_per_file)
                        # round the end time to the nearest second as the timestamp is only accurate to the second
                        end = end.replace(microsecond=0)
                        # set the times for the next files
                        start = end
            else:
                day_df = self.no_jitter(day_df)

            # drop any rows with duplicate uri times, keeping the first
            # duplicates can be caused by the jitter correction
            if "uri" in day_df.columns:
                day_df = day_df.drop_duplicates(subset=["uri"], keep="first")

            # save explicitly as UTC by setting the timezone in the start and end times
            day_df["start"] = day_df["start"].dt.tz_localize("UTC")
            day_df["end"] = day_df["end"].dt.tz_localize("UTC")

            self.save_day(self.day, day_df)

        except Exception as e:
            self.log.exception(f"Error correcting metadata for  {self.day}. {e}")
        finally:
            self.log.debug(
                f"Done correcting metadata for {self.day}. Saved to {self.json_base_dir}"
            )

    def no_jitter(self, day_df: pd.DataFrame) -> pd.DataFrame:
        """
        Set the jitter to 0 and calculate the end time from the start time and the duration
        :param day_df:
            The dataframe to correct
        :return:
            The corrected dataframe
        """
        self.log.warning(
            f"Cannot correct {self.day}. Using file start times as is, setting jitter to 0 and using "
            f"calculated end times."
        )
        # calculate the difference between each row start time and save as diff in a copy of the dataframe
        day_df = day_df.copy()
        day_df["diff"] = day_df["start"].diff()
        day_df["jitter_secs"] = 0
        # calculate the end time which is the start time plus the number of seconds in the file
        day_df["end"] = day_df["start"] + pd.to_timedelta(
            day_df["duration_secs"], unit="s"
        )
        return day_df

    def save_day(self, day: datetime.datetime, day_df: pd.DataFrame, prefix: str = ""):
        """
        Save the day's metadata to a single json file either locally or to s3
        :param day:
            The day to save
        :param day_df:
            The dataframe containing the metadata for the day
        :param prefix:
            An optional prefix for the filename
        :return:
        """
        # if the exception column is full of empty strings, then drop it
        if "exception" in day_df.columns and day_df["exception"].str.len().sum() == 0:
            day_df.drop(columns=["exception"], inplace=True)

        # drop the pcm, fs, subtype, etc. columns
        day_df.drop(columns=["fs", "subtype", "jitter_secs"], inplace=True)

        # if there is a diff column, then drop it
        if "diff" in day_df.columns:
            day_df.drop(columns=["diff"], inplace=True)

        # Save with second accuracy to a temporary file formatted with ISO date format
        df_final = day_df.sort_values(by=["start"])

        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            if prefix:
                temp_metadata = tmp_path / f"{prefix}_{day:%Y%m%d}.json"
            else:
                temp_metadata = tmp_path / f"{day:%Y%m%d}.json"

            df_final.to_json(
                temp_metadata.as_posix(),
                orient="records",
                date_format="iso",
                date_unit="s",
            )
            self.log.debug(f"Wrote {temp_metadata.as_posix()}")

            # read the file back in using records format with json
            with open(temp_metadata.as_posix(), "r") as f:
                dict_records = json.load(f)

            # write the file back out with indenting
            with open(temp_metadata.as_posix(), "w", encoding="utf-8") as f:
                json.dump(dict_records, f, ensure_ascii=True, indent=4)

            # copy the file to a local metadata directory with year subdirectory
            output_path = Path(self.json_base_dir, str(day.year))
            output_path.mkdir(parents=True, exist_ok=True)
            shutil.copy2(temp_metadata.as_posix(), output_path)
            self.log.info(f"Wrote {output_path}/{temp_metadata.name}")



from argparse import ArgumentParser, RawTextHelpFormatter

from pbp.plot_const import (
    DEFAULT_DPI,
    DEFAULT_LAT_LON_FOR_SOLPOS,
    DEFAULT_TITLE,
    DEFAULT_YLIM,
    DEFAULT_CMLIM,
)
from pbp import get_pbp_version


def parse_arguments():
    description = "Generate summary plots for given netcdf files."

    parser = ArgumentParser(description=description, formatter_class=RawTextHelpFormatter)

    parser.add_argument(
        "--version",
        action="version",
        version=get_pbp_version(),
    )

    parser.add_argument(
        "--latlon",
        type=float,
        nargs=2,
        default=DEFAULT_LAT_LON_FOR_SOLPOS,
        metavar=("lat", "lon"),
        help="Lat/Lon for solar position calculation . Default: %(default)s",
    )

    parser.add_argument(
        "--title",
        type=str,
        default=f"'{DEFAULT_TITLE}'",
        metavar="string",
        help="Title for the plot. Default: %(default)s",
    )

    parser.add_argument(
        "--ylim",
        type=int,
        nargs=2,
        default=DEFAULT_YLIM,
        metavar=("lower", "upper"),
        help="Limits for the y-axis. Default: %(default)s",
    )

    parser.add_argument(
        "--cmlim",
        type=int,
        nargs=2,
        default=DEFAULT_CMLIM,
        metavar=("vmin", "vmax"),
        help="Parameters passed to pcolormesh. Default: %(default)s",
    )

    parser.add_argument(
        "--dpi",
        type=float,
        default=DEFAULT_DPI,
        metavar="value",
        help="DPI to use for the plot. Default: %(default)s",
    )

    parser.add_argument(
        "--show",
        default=False,
        action="store_true",
        help="Also show the plot",
    )

    parser.add_argument(
        "--only-show",
        default=False,
        action="store_true",
        help="Only show the plot (do not generate .jpg files)",
    )

    parser.add_argument(
        "--engine",
        type=str,
        metavar="name",
        default="h5netcdf",
        help="Engine given to xarray.open_dataset. Default: %(default)s",
    )

    parser.add_argument("netcdf", nargs="+", help="netcdf file(s) to plot")

    return parser.parse_args()


def main():
    opts = parse_arguments()
    # pylint: disable=import-outside-toplevel
    import xarray as xr

    from pbp.plotting import plot_dataset_summary

    show = opts.show or opts.only_show
    for nc_filename in opts.netcdf:
        print(f"plotting {nc_filename} at {opts.dpi} dpi")
        ds = xr.open_dataset(nc_filename, engine=opts.engine)
        jpeg_filename = None if opts.only_show else nc_filename.replace(".nc", ".jpg")
        plot_dataset_summary(
            ds,
            lat_lon_for_solpos=opts.latlon,
            title=opts.title,
            ylim=opts.ylim,
            cmlim=opts.cmlim,
            dpi=opts.dpi,
            jpeg_filename=jpeg_filename,
            show=show,
        )
        if jpeg_filename is not None:
            print(f"   done: {jpeg_filename}")


if __name__ == "__main__":
    main()



from pbp.process_helper import ProcessHelper, ProcessDayResult
from pbp.file_helper import FileHelper
from pbp.logging_helper import create_logger
from typing import Optional
from botocore.client import BaseClient
from pbp.plotting import plot_dataset_summary
from pbp import get_pbp_version

from google.cloud.storage import Client as GsClient

import xarray as xr
import pandas as pd
import matplotlib.pyplot as plt


class HmbGen:
    def __init__(self) -> None:
        """
        Create a new object for HMB generation.
        Use the various set methods to set the parameters,
        then call process_date() to execute the process for a given date.
        """
        self._json_base_dir: str = ""
        self._global_attrs_uri: str = ""
        self._variable_attrs_uri: str = ""
        self._voltage_multiplier: float = 1.0
        self._sensitivity: float | str = 1.0
        self._subset_to: Optional[tuple[int, int]] = None
        self._download_dir: str = ""
        self._output_dir: str = ""
        self._output_prefix: str = ""

        self._assume_downloaded_files: bool = False
        self._retain_downloaded_files: bool = False
        self._print_downloading_lines: bool = False

        # Other attributes
        self._s3_client: Optional[BaseClient] = None
        self._gs_client: Optional[GsClient] = None

        self._hmb_gen: Optional[_HmbGen] = None

    def set_json_base_dir(self, json_base_dir: str) -> None:
        """
        Set the base directory where JSON files are located.
        """
        self._json_base_dir = json_base_dir

    def set_global_attrs_uri(self, global_attrs_uri: str) -> None:
        """
        Set the URI for global attributes.
        """
        self._global_attrs_uri = global_attrs_uri

    def set_variable_attrs_uri(self, variable_attrs_uri: str) -> None:
        """
        set the URI for variable attributes.
        """
        self._variable_attrs_uri = variable_attrs_uri

    def set_voltage_multiplier(self, voltage_multiplier: float) -> None:
        """
        Set the voltage multiplier. By default, this is 1.0.
        """
        self._voltage_multiplier = voltage_multiplier

    def set_sensitivity(self, sensitivity: float | str) -> None:
        """
        Set Sensitivity flat value (a float), or URI of sensitivity file.
        """
        self._sensitivity = sensitivity

    def set_subset_to(self, subset_to: tuple[int, int]) -> None:
        """
        Set the frequency subset to use for the PSD.

        :param subset_to:
            Tuple of (lower, upper), with lower inclusive, upper exclusive.
        """
        self._subset_to = subset_to

    def set_download_dir(self, download_dir: str) -> None:
        """
        Set the download directory.
        """
        self._download_dir = download_dir

    def set_output_dir(self, output_dir: str) -> None:
        """
        Set the output directory.
        """
        self._output_dir = output_dir

    def set_output_prefix(self, output_prefix: str) -> None:
        """
        Set the output prefix.
        """
        self._output_prefix = output_prefix

    def set_assume_downloaded_files(self, value: bool) -> None:
        """
        :param value:
            If True, skip downloading files that already exist in the download directory.
        """
        self._assume_downloaded_files = value

    def set_retain_downloaded_files(self, value: bool) -> None:
        """
        :param value:
            If True, do not remove downloaded files after use.
        """
        self._retain_downloaded_files = value

    def set_print_downloading_lines(self, value: bool) -> None:
        """
        :param value:
            If True, print "downloading <uri>" lines to console.
        """
        self._print_downloading_lines = value

    def set_s3_client(self, s3_client: BaseClient) -> None:
        """
        Set the S3 client.
        """
        if self._gs_client:
            raise ValueError("A GS client has already been set.")
        self._s3_client = s3_client

    def set_gs_client(self, gs_client: GsClient) -> None:
        """
        Set the GS client.
        """
        if self._s3_client:
            raise ValueError("A S3 client has already been set.")
        self._gs_client = gs_client

    def check_parameters(self) -> str | None:
        """
        Performs a basic check of the parameters,
        especially verifying that the required ones are given.
        Call this before performing any processing.

        :return: None if all looks OK, or an error message.
        """
        errors = []

        if not self._json_base_dir:
            errors.append("- json_base_dir not set")

        if not self._global_attrs_uri:
            errors.append("- global_attrs_uri not set")

        if not self._variable_attrs_uri:
            errors.append("- variable_attrs_uri not set")

        if not self._download_dir:
            errors.append("- download_dir not set")

        if not self._output_dir:
            errors.append("- output_dir not set")

        if not self._subset_to:
            errors.append("- subset_to not set")

        if not self._sensitivity:
            errors.append("- sensitivity not set.")
        elif not isinstance(self._sensitivity, (float, int, str)):
            errors.append("- sensitivity must be a number or a string")

        if not isinstance(self._subset_to, tuple):
            errors.append("- subset_to must be a tuple")
        else:
            if len(list(self._subset_to)) != 2:
                errors.append("- subset_to must be a tuple of length 2")
            if not isinstance(self._subset_to[0], int) or not isinstance(
                self._subset_to[1], int
            ):
                errors.append("- subset_to must contain integers")

        if not self._s3_client and not self._gs_client:
            errors.append("- No S3 or GS client has been set")

        if len(errors) > 0:
            return "\n".join(errors)

        # make mypy happy
        assert isinstance(self._subset_to, tuple)

        self._hmb_gen = _HmbGen(
            json_base_dir=self._json_base_dir,
            global_attrs_uri=self._global_attrs_uri,
            variable_attrs_uri=self._variable_attrs_uri,
            voltage_multiplier=self._voltage_multiplier,
            sensitivity=self._sensitivity,
            subset_to=self._subset_to,
            download_dir=self._download_dir,
            output_dir=self._output_dir,
            output_prefix=self._output_prefix,
            assume_downloaded_files=self._assume_downloaded_files,
            retain_downloaded_files=self._retain_downloaded_files,
            print_downloading_lines=self._print_downloading_lines,
            s3_client=self._s3_client,
            gs_client=self._gs_client,
        )
        return None

    def process_date(self, date: str) -> ProcessDayResult | str:
        """
        Generates NetCDF file with the result of processing all segments of the given day.

        :param date:
            Date to process in YYYYMMDD format.
        :return:
            ProcessDayResult, or a  string indicating any errors
             if no segments at all were processed for the day.
        """
        if not self._hmb_gen:
            return "Missing or invalid parameters. Call check_parameters() first."

        result = self._hmb_gen.process_date(date)
        return result or f"No segments processed for {date}."

    def plot_date(
        self,
        date: str,
        lat_lon_for_solpos: tuple[float, float],
        title: str,
        ylim: tuple[int, int],
        cmlim: tuple[int, int],
        dpi: int,
        show: bool = False,
    ) -> None:
        """
        Generate a summary plot for the given date.
        Make sure the NetCDF file for the given date has been generated first.
        The output will be saved as a JPEG file with name derived from the input date
        resulting in a sibling file to the NetCDF file.

        :param date: Date to plot in YYYYMMDD format.
        :param lat_lon_for_solpos: Lat/Lon for solar position calculation.
        :param title: Title for the plot.
        :param ylim: Limits for the y-axis.
        :param cmlim: Limits passed to pcolormesh.
        :param dpi: DPI to use for the plot.
        :param show: Whether to also show the plot. By default, only the JPEG file is generated.
        """
        if not self._hmb_gen:
            raise ValueError(
                "Missing or invalid parameters. Call check_parameters() first."
            )

        return self._hmb_gen.plot_date(
            date,
            lat_lon_for_solpos,
            title,
            ylim,
            cmlim,
            dpi,
            show,
        )


def _version() -> str:
    return f"pbp v{get_pbp_version()}: "


class _HmbGen:
    def __init__(
        self,
        json_base_dir: str,
        global_attrs_uri: str,
        variable_attrs_uri: str,
        voltage_multiplier: float,
        sensitivity: float | str,
        subset_to: tuple[int, int],
        download_dir: str,
        output_dir: str,
        output_prefix: str,
        assume_downloaded_files: bool,
        retain_downloaded_files: bool,
        print_downloading_lines: bool,
        s3_client: Optional[BaseClient],
        gs_client: Optional[GsClient],
    ) -> None:
        self.json_base_dir = json_base_dir
        self.global_attrs_uri = global_attrs_uri
        self.variable_attrs_uri = variable_attrs_uri
        self.voltage_multiplier = voltage_multiplier
        self.subset_to = subset_to

        self.sensitivity_uri: Optional[str] = None
        self.sensitivity_flat_value: Optional[float] = None
        if isinstance(sensitivity, str):
            self.sensitivity_uri = sensitivity
        else:
            self.sensitivity_flat_value = sensitivity

        self.download_dir = download_dir
        self.output_dir = output_dir
        self.output_prefix = output_prefix

        self.assume_downloaded_files = assume_downloaded_files
        self.retain_downloaded_files = retain_downloaded_files
        self.print_downloading_lines = print_downloading_lines

        # Other attributes
        self.s3_client = s3_client
        self.gs_client = gs_client

        # -----------------------------------------------------
        # Reset plotting overrides set in pypam
        # (these cause the PBP plot_dataset_summary function to fail
        # https://github.com/mbari-org/pbp/issues/21#issuecomment-2261642486)
        plt.rcParams.update({"text.usetex": False})
        pd.plotting.deregister_matplotlib_converters()

    def process_date(self, date: str) -> Optional[ProcessDayResult]:
        nc_filename = f"{self.output_dir}/{self.output_prefix}{date}.nc"

        print(f"{_version()}: Processing {date} to generate {nc_filename}...")

        log_filename = nc_filename.replace(".nc", ".log")
        log = create_logger(
            log_filename_and_level=(log_filename, "INFO"),
            console_level=None,
        )

        file_helper = FileHelper(
            log=log,
            json_base_dir=self.json_base_dir,
            s3_client=self.s3_client,
            gs_client=self.gs_client,
            download_dir=self.download_dir,
            assume_downloaded_files=self.assume_downloaded_files,
            retain_downloaded_files=self.retain_downloaded_files,
            print_downloading_lines=self.print_downloading_lines,
        )

        process_helper = ProcessHelper(
            log=log,
            file_helper=file_helper,
            output_dir=self.output_dir,
            output_prefix=self.output_prefix,
            global_attrs_uri=self.global_attrs_uri,
            variable_attrs_uri=self.variable_attrs_uri,
            voltage_multiplier=self.voltage_multiplier,
            sensitivity_uri=self.sensitivity_uri,
            sensitivity_flat_value=self.sensitivity_flat_value,
            subset_to=self.subset_to,
        )

        # now, get the HMB result:
        print(f"::: Started processing {date=}")
        result = process_helper.process_day(date)
        print(f":::   Ended processing {date=} =>  {nc_filename=}")
        return result

    def plot_date(
        self,
        date: str,
        lat_lon_for_solpos: tuple[float, float],
        title: str,
        ylim: tuple[int, int],
        cmlim: tuple[int, int],
        dpi: int,
        show: bool = False,
    ) -> None:
        nc_filename = f"{self.output_dir}/{self.output_prefix}{date}.nc"
        jpeg_filename = nc_filename.replace(".nc", ".jpg")

        print(f"{_version()}: Generating plot for {date} {jpeg_filename}...")

        ds = xr.open_dataset(nc_filename, engine="h5netcdf")
        plot_dataset_summary(
            ds,
            lat_lon_for_solpos=lat_lon_for_solpos,
            title=title,
            ylim=ylim,
            cmlim=cmlim,
            dpi=dpi,
            jpeg_filename=jpeg_filename,
            show=show,
        )
        print(f"   done: {jpeg_filename}")



from typing import Any, Generator, Tuple, Union

import numpy as np


def parse_date(date: str) -> Tuple[int, int, int]:
    """
    Parses given string into a (year, month, day) integer tuple.
    :param date:
        Digit string with YYYYMMDD format.
    :return:
        (year, month, day) tuple.
    """
    assert date.isdigit() and len(date) == 8
    year, month, day = int(date[:4]), int(date[4:6]), int(date[6:8])
    return year, month, day


def gen_hour_minute_times(
    segment_size_in_mins: int = 1,
) -> Generator[Tuple[int, int], None, None]:
    """
    Generate a sequence of starting (hour, minute) tuples to cover a whole day.
    For example, for segment_size_in_mins=1, the sequence is:
    (0, 0), ..., (23, 58), (23, 59)
    """
    day_minutes = 24 * 60
    current_minute = 0

    while current_minute < day_minutes:
        at_hour, at_minute = divmod(current_minute, 60)
        yield at_hour, at_minute
        current_minute += segment_size_in_mins


def map_prefix(prefix_map: str, s: str) -> str:
    """
    Helper to replace a prefix to another prefix in given string
    according to prefix_map.
    :param prefix_map:  Like "old~new".
    :param s:  The string to replace the prefix in.
    :return:  Resulting string, possibly unchanged.
    """
    if "~" in prefix_map:
        old, new = prefix_map.split("~", 2)
        if s.startswith(old):
            return new + s[len(old) :]
    return s


# Using Union as `X | Y syntax for unions requires Python 3.10`
def brief_list(lst: Union[list, np.ndarray[Any, Any]], max_items: int = 6) -> str:
    """
    Helper to format a list as a string, with a maximum number of items.
    :param lst:  The list to format.
    :param max_items:  The maximum number of items to show.
    :return:  The formatted string.
    """
    if len(lst) > max_items:
        half = max_items // 2
        rest = max_items - half
        prefix = ", ".join(str(x) for x in lst[:half])
        postfix = ", ".join(str(x) for x in lst[-rest:])
        return f"[{prefix}, ..., {postfix}]"
    return str(lst)



# pbp, Apache License 2.0
# Filename: json_generator/gen_soundtrap.py
# Description:  Captures SoundTrap metadata either from a local directory of S3 bucket
import urllib
from typing import List

import boto3
import datetime
import pandas as pd
import re
import pytz

from datetime import timedelta
from pathlib import Path

from progressbar import progressbar

from pbp.json_generator.gen_abstract import MetadataGeneratorAbstract
from pbp.json_generator.metadata_extractor import SoundTrapWavFile
from pbp.json_generator.corrector import MetadataCorrector
from pbp.json_generator.utils import parse_s3_or_gcp_url, InstrumentType


class SoundTrapMetadataGenerator(MetadataGeneratorAbstract):
    """
    Captures SoundTrap wav file metadata either from a local directory or S3 bucket.
    """

    START = datetime.datetime.now(pytz.utc)
    END = datetime.datetime.now(pytz.utc)

    def __init__(
        self,
        log,  # : loguru.Logger,
        uri: str,
        json_base_dir: str,
        prefix: List[str],
        start: datetime.datetime = START,
        end: datetime.datetime = END,
    ):
        """
        :param uri:
            The local directory or S3 bucket that contains the wav files
        :param json_base_dir:
            The local directory to write the json files to
        :param prefix:
            The search pattern to match the wav files, e.g. 'MARS'
        :param start:
            The start date to search for wav files
        :param end:
            The end date to search for wav files check is done.
        :return:
        """
        super().__init__(log, uri, json_base_dir, prefix, start, end, 0.0)

    def run(self):
        try:
            xml_cache_path = Path(self.json_base_dir) / "xml_cache"
            xml_cache_path.mkdir(exist_ok=True, parents=True)
            wav_files = []

            self.log.info(
                f"Searching in {self.audio_loc}/*.wav for wav files that match the prefix {self.prefix}* ..."
            )

            bucket, prefix, scheme = parse_s3_or_gcp_url(self.audio_loc)
            # This does not work for GCS
            if scheme == "gs":
                self.log.error("GS not supported for SoundTrap")
                return

            # Set the start and end dates to 1 day before and after the start and end dates
            start_dt = self.start - timedelta(days=1)
            end_dt = self.end + timedelta(days=1)

            def get_file_date(xml_file: str) -> datetime:
                """
                Check if the xml file is in the search pattern and is within the start and end dates
                :param xml_file:
                    The xml file with the metadata
                :return:
                    Record starting datetime if the file is within the start and end dates; otherwise, return None
                """
                xml_file_path = Path(xml_file)
                # see if the file is a regexp match to self.prefix
                for s in self.prefix:
                    rc = re.search(s, xml_file_path.stem)

                    if rc and rc.group(0):
                        try:
                            pattern_date1 = re.compile(
                                r"(\d{4})(\d{2})(\d{2})_(\d{2})(\d{2})(\d{2})"
                            )  # 20161025T184500Z
                            search = pattern_date1.search(xml_file_path.stem)
                            if search:
                                match = search.groups()
                                year, month, day, hour, minute, second = map(int, match)
                                f_path_dt = datetime.datetime(
                                    year, month, day, hour, minute, second
                                )
                            else:
                                f_path_dt = datetime.datetime.strptime(
                                    xml_file_path.stem.split(".")[1], "%y%m%d%H%M%S"
                                )
                            if start_dt <= f_path_dt <= end_dt:
                                return f_path_dt
                        except ValueError:
                            self.log.error(f"Could not parse {xml_file_path.name}")
                return None

            if scheme == "file":
                parsed_uri = urllib.parse.urlparse(self.audio_loc)
                wav_path = Path(parsed_uri.path)
                for filename in progressbar(
                    sorted(wav_path.rglob("*.xml")), prefix="Searching : "
                ):
                    wav_path = filename.parent / f"{filename.stem}.wav"
                    start_dt = get_file_date(filename)
                    if start_dt:
                        wav_files.append(
                            SoundTrapWavFile(wav_path.as_posix(), filename, start_dt)
                        )
            else:
                # if the audio_loc is a s3 url, then we need to list the files in buckets that cover the start and end
                # dates
                self.log.info(f"Searching between {start_dt} and {end_dt}")

                client = boto3.client("s3")
                paginator = client.get_paginator("list_objects")

                operation_parameters = {"Bucket": bucket}
                page_iterator = paginator.paginate(**operation_parameters)
                self.log.info(
                    f"Searching in bucket: {bucket} for .wav and .xml files between {start_dt} and {end_dt} "
                )
                # list the objects in the bucket
                # loop through the objects and check if they match the search pattern
                for page in page_iterator:
                    for obj in page["Contents"]:
                        key = obj["Key"]

                        if ".xml" in key and get_file_date(key):
                            xml_path = xml_cache_path / key
                            wav_uri = f"s3://{bucket}/{key}".replace(
                                "self.log.xml", "wav"
                            )

                            # Check if the xml file is in the cache directory
                            if not xml_path.exists():
                                # Download the xml file to a temporary directory
                                self.log.info(f"Downloading {key} ...")
                                client.download_file(bucket, key, xml_path)

                            start_dt = get_file_date(wav_uri)
                            if start_dt:
                                wav_files.append(
                                    SoundTrapWavFile(wav_uri, xml_path, start_dt)
                                )

            self.log.info(
                f"Found {len(wav_files)} files to process that cover the period {start_dt} - {end_dt}"
            )

            if len(wav_files) == 0:
                return

            # sort the files by start time
            wav_files.sort(key=lambda x: x.start)

            # create a dataframe from the wav files
            self.log.info(
                f"Creating dataframe from {len(wav_files)} files spanning {wav_files[0].start} to {wav_files[-1].start}..."
            )
            for wc in wav_files:
                df_wav = wc.to_df()

                # concatenate the metadata to the dataframe
                self.df = pd.concat([self.df, df_wav], axis=0)

            # drop any rows with duplicate uris, keeping the first
            self.df = self.df.drop_duplicates(subset=["uri"], keep="first")

        except Exception as ex:
            self.log.exception(str(ex))
        finally:
            days = (self.end - self.start).days + 1

            if len(self.df) == 0:
                self.log.info(f"No data found between {self.start} and {self.end}")
                return

            # Correct the metadata for each day
            for day in range(days):
                self.log.debug(f"Running metadata corrector for {day}")
                corrector = MetadataCorrector(
                    self.log,
                    self.df,
                    self.json_base_dir,
                    self.start + timedelta(days=day),
                    InstrumentType.NRS,
                    False,
                )
                corrector.run()


if __name__ == "__main__":
    from pbp.logging_helper import create_logger

    log_dir = Path("tests/log")
    json_dir = Path("tests/json/soundtrap")
    log_dir.mkdir(exist_ok=True, parents=True)
    json_dir.mkdir(exist_ok=True, parents=True)

    log = create_logger(
        log_filename_and_level=(
            f"{log_dir}/test_soundtrap_metadata_generator.log",
            "INFO",
        ),
        console_level="INFO",
    )

    start = datetime.datetime(2023, 7, 18)
    end = datetime.datetime(2023, 7, 19)
    gen = SoundTrapMetadataGenerator(
        log, "s3://pacific-sound-ch01", json_dir.as_posix(), ["7000"], start, end
    )
    gen.run()



# pbp, Apache License 2.0
# Filename: metadata/generator/gen_nrs.py
# Description:  Captures NRS recorded metadata in a pandas dataframe from either a local directory or gs bucket.

import re
from datetime import timedelta, datetime
import time
from typing import List

from google.cloud import storage

import pandas as pd
from pathlib import Path
from progressbar import progressbar
from pbp.json_generator.corrector import MetadataCorrector
from pbp.json_generator.metadata_extractor import FlacFile, GenericWavFile as WavFile
from pbp.json_generator.gen_abstract import MetadataGeneratorAbstract
from pbp.json_generator.utils import parse_s3_or_gcp_url, InstrumentType


class NRSMetadataGenerator(MetadataGeneratorAbstract):
    def __init__(
        self,
        log,  # : loguru.Logger,
        uri: str,
        json_base_dir: str,
        start: datetime,
        end: datetime,
        prefix: List[str],
        seconds_per_file: float = 14400.0,
    ):
        """
        Captures NRS audio metadata in a pandas dataframe from either a local directory or GS bucket.
        :param uri:
            The local directory or GCP bucket that contains the audio files
        :param json_base_dir:
            The local directory to store the metadata
        :param start:
            The start date to search for flac files
        :param end:
            The end date to search for flac files
        :param prefix:
            The search pattern to match the flac files, e.g. 'MARS' for MARS_YYYYMMDD_HHMMSS.flac
        :param seconds_per_file:
            The number of seconds per file expected in a flac/wav file to check for missing data. If 0, then no check is done.
        :return:
        """
        super().__init__(log, uri, json_base_dir, prefix, start, end, seconds_per_file)

    def run(self):
        self.log.info(f"Generating metadata for {self.start} to {self.end}...")

        bucket, prefix, scheme = parse_s3_or_gcp_url(self.audio_loc)

        # S3 is not supported for NRS
        if scheme == "s3":
            self.log.error("S3 is not supported for NRS audio files")
            return

        def parse_filename(f: str) -> datetime or None:
            """
            Check if the file matches the search pattern and is within the start and end dates
            :param f:
                The path to the file
            :return: The beginning recording time of the file
            """
            f_path = Path(f)
            f_path_dt = None

            for s in self.prefix:
                # see if the file is a regexp match to search
                rc = re.search(s, f_path.stem)

                if rc and rc.group(0):
                    try:
                        pattern_date = re.compile(
                            r"(\d{4})(\d{2})(\d{2})_(\d{2})(\d{2})(\d{2})"
                        )  # 20191231_230836
                        search = pattern_date.search(f_path.stem)
                        if search:
                            match = search.groups()
                            year, month, day, hour, minute, second = map(int, match)
                            if second == 60:  # this is a bug in the flac files names
                                second = 59
                            f_path_dt = datetime(year, month, day, hour, minute, second)
                            return f_path_dt
                        else:
                            self.log.error(f"Could not parse {f_path.name}")
                            return None
                    except ValueError:
                        self.log.error(f"Could not parse {f_path.name}")
                        return None

            return f_path_dt

        sound_files = []
        self.df = None
        self.log.info(
            f"Searching in {self.audio_loc}/ for files that match the search pattern {self.prefix}* ..."
        )

        # set the window to 1 flac file to account for any missing data
        minutes_window = int(self.seconds_per_file / 60)

        # pad the start and end dates to account for any missing data
        start_dt = self.start - timedelta(minutes=minutes_window)
        end_dt = self.end + timedelta(days=1)

        if scheme == "file" or scheme == "":
            sound_path = Path(f"/{bucket}/{prefix}")
            # First search for flac files
            for filename in progressbar(
                sorted(sound_path.rglob("*.flac")), prefix="Searching : "
            ):
                flac_dt = parse_filename(filename)
                if start_dt <= flac_dt <= end_dt:
                    self.log.info(f"Found file {filename} with timestamp {flac_dt}")
                    sound_files.append(FlacFile(self.log, str(filename), flac_dt))
            # Next search for wav files
            for filename in progressbar(
                sorted(sound_path.rglob("*.wav")), prefix="Searching : "
            ):
                wav_dt = parse_filename(filename)
                if start_dt <= wav_dt <= end_dt:
                    self.log.info(f"Found file {filename} with timestamp {wav_dt}")
                    sound_files.append(WavFile(self.log, str(filename), wav_dt))

        if scheme == "gs":
            client = storage.Client.create_anonymous_client()
            bucket_obj = client.get_bucket(bucket)

            # get list of files - this is a generator
            # data is organized in a flat filesystem, so there are no optimizations here for querying
            blobs = bucket_obj.list_blobs(prefix=prefix)
            for i, blob in enumerate(blobs):
                self.log.info(f"Processing {blob.name}")
                f_path = f"gs://{bucket}/{blob.name}"
                f_dt = parse_filename(f_path)
                if start_dt <= f_dt <= end_dt:
                    self.log.info(f"Found file {blob.name} with timestamp {f_dt}")
                    sound_files.append(FlacFile(self.log, f_path, f_dt))
                # delay to avoid 400 error
                if i % 100 == 0:
                    self.log.info(f"{i} files processed")
                    time.sleep(1)
                if f_dt > end_dt:
                    break

        self.log.info(
            f"Found {len(sound_files)} files to process that cover the period {start_dt} - {end_dt}"
        )

        if len(sound_files) == 0:
            return

        # sort the files by start time
        sound_files.sort(key=lambda x: x.start)
        for wc in sound_files:
            df_flac = wc.to_df()

            # concatenate the metadata to the dataframe
            self.df = pd.concat([self.df, df_flac], axis=0)

        # correct each day in the range
        for day in pd.date_range(self.start, self.end, freq="D"):
            try:
                # create a dataframe from the flac files
                self.log.info(
                    f"Creating dataframe from {len(sound_files)} "
                    f"files spanning {sound_files[0].start} to {sound_files[-1].start} in self.json_base_dir..."
                )

                self.log.debug(f" Running metadata corrector for {day}")
                corrector = MetadataCorrector(
                    self.log,
                    self.df,
                    self.json_base_dir,
                    day,
                    InstrumentType.NRS,
                    False,
                    self.seconds_per_file,
                )
                corrector.run()

            except Exception as ex:
                self.log.exception(str(ex))


if __name__ == "__main__":
    from pbp.logging_helper import create_logger

    log_dir = Path("tests/log")
    json_dir = Path("tests/json/nrs")
    log_dir.mkdir(exist_ok=True, parents=True)
    json_dir.mkdir(exist_ok=True, parents=True)

    log = create_logger(
        log_filename_and_level=(
            f"{log_dir}/test_nrs_metadata_generator.log",
            "INFO",
        ),
        console_level="INFO",
    )

    start = datetime(2019, 10, 24, 0, 0, 0)
    end = datetime(2019, 11, 1, 0, 0, 0)

    generator = NRSMetadataGenerator(
        log,
        uri="gs://noaa-passive-bioacoustic/nrs/audio/11/nrs_11_2019-2021/audio",
        json_base_dir=json_dir.as_posix(),
        prefix=["NRS11"],
        start=start,
        end=end,
    )
    generator.run()



# pbp, Apache License 2.0
# Filename: metadata/generator/gen_abstract.py
# Description:  Abstract class that captures sound wav metadata
from datetime import datetime
from typing import List

import pandas as pd


class MetadataGeneratorAbstract(object):
    def __init__(
        self,
        log,  # : loguru.Logger,
        audio_loc: str,
        json_base_dir: str,
        prefix: List[str],
        start: datetime,
        end: datetime,
        seconds_per_file: float = 0.0,
        **kwargs,
    ):
        """
        Abstract class for capturing sound wav metadata
        :param audio_loc:
            The local directory or cloud bucket that contains the wav files
        :param json_base_dir:
            The local directory to write the json files to
        :param prefix:
            The search pattern to match the wav files, e.g. 'MARS'
        :param start:
            The start date to search for wav files
        :param end:
            The end date to search for wav files
        :param seconds_per_file:
            The number of seconds per file expected in a wav file to check for missing data. If missing, then no check is done.
        :return:
        """
        try:
            self.audio_loc = audio_loc
            self.json_base_dir = json_base_dir
            self.df = pd.DataFrame()
            self.start = start
            self.end = end
            self.prefix = prefix
            self._log = log
            self._seconds_per_file = None if seconds_per_file == 0 else seconds_per_file
        except Exception as e:
            raise e

    @property
    def seconds_per_file(self):
        return self._seconds_per_file

    @property
    def log(self):
        return self._log

    # abstract run method
    def run(self):
        pass



import json
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Generator, List
from urllib.parse import urlparse
from dataclasses_json import config, dataclass_json
from dateutil import parser as iso8601_parser


@dataclass_json
@dataclass
class JEntry:
    """
    Captures each object in the array contained in the given Json file.
    """

    uri: str
    duration_secs: float
    start: datetime = field(
        metadata=config(
            encoder=lambda dt: dt.strftime("%Y-%m-%dT%H:%M:%S.000Z"),
            decoder=iso8601_parser.parse,
        )
    )

    @property
    def path(self) -> str:
        return urlparse(self.uri).path


def parse_json_contents(contents: str) -> Generator[JEntry, None, None]:
    reported_uris = set()
    for item in json.loads(contents):
        entry = JEntry.from_dict(item)  # type: ignore [attr-defined]
        if entry.uri in reported_uris:
            print(f"warning: Skipping duplicate json entry: uri={entry.uri}")
            continue
        reported_uris.add(entry.uri)
        yield entry


def parse_json_file(filename: str) -> Generator[JEntry, None, None]:
    with open(filename, "r", encoding="UTF-8") as f:
        return parse_json_contents(f.read())


@dataclass_json
@dataclass
class JEntryIntersection:
    entry: JEntry

    # relative to the start of the JEntry
    start_secs: int
    duration_secs: int


def get_intersecting_entries(
    log,  # : loguru.Logger,
    json_entries: List[JEntry],
    year: int,
    month: int,
    day: int,
    at_hour: int,
    at_minute: int,
    segment_size_in_mins: int = 1,
) -> List[JEntryIntersection]:
    """
    Gets the list of intersecting entries for the UTC "start minute"
    given by (year, month, day, at_hour, at_minute).

    :param log:
        Logger
    :param json_entries:
        JSON entries for the day
    :param year:
        Year associated to the start minute
    :param month:
        Month associated to the start minute
    :param day:
        Day associated to the start minute
    :param at_hour:
        Hour associated to the start minute
    :param at_minute:
        Minute associated to the start minute
    :param segment_size_in_mins:
        The size of the segment in minutes, by default 1.

    :return:
        The list of intersecting entries
    """
    # for logging purposes:
    time_spec = (
        f"year={year} month={month} day={day} at_hour={at_hour} at_minute={at_minute}"
    )
    log.debug(f"get_intersecting_entries: {time_spec} {len(json_entries)=}")

    # the requested start minute as datetime:
    dt = datetime(year, month, day, at_hour, at_minute, tzinfo=timezone.utc)
    # the start of the requested start minute in seconds:
    minute_start_in_secs: int = int(dt.timestamp())
    # the end of the requested start minute in seconds:
    minute_end_in_secs: int = minute_start_in_secs + segment_size_in_mins * 60

    intersecting_entries: List[JEntryIntersection] = []
    tot_duration_secs = 0
    for entry in json_entries:
        entry_start_in_secs: int = int(entry.start.timestamp())
        entry_end_in_secs: int = entry_start_in_secs + int(entry.duration_secs)
        if (
            entry_start_in_secs <= minute_end_in_secs
            and entry_end_in_secs >= minute_start_in_secs
        ):
            start_secs = (
                max(entry_start_in_secs, minute_start_in_secs) - entry_start_in_secs
            )
            end_secs = min(entry_end_in_secs, minute_end_in_secs) - entry_start_in_secs
            duration_secs = end_secs - start_secs
            intersecting_entries.append(
                JEntryIntersection(entry, start_secs, duration_secs)
            )
            tot_duration_secs += duration_secs

    warning = 0 == len(intersecting_entries)

    def log_msg():
        uris = [i.entry.uri for i in intersecting_entries]
        uris_str = "\n  ".join([f"[{e}] {uri}" for e, uri in enumerate(uris)])
        return f"{time_spec}: intersection uris({len(uris)}):\n  {uris_str}"

    if warning:
        log.opt(lazy=True).warning("get_intersecting_entries: {}", log_msg)
    else:
        log.opt(lazy=True).debug("get_intersecting_entries: {}", log_msg)

    return intersecting_entries



from typing import Optional

import matplotlib.dates as md
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pvlib
import xarray as xr
from matplotlib import gridspec

from pbp.plot_const import (
    DEFAULT_DPI,
    DEFAULT_LAT_LON_FOR_SOLPOS,
    DEFAULT_TITLE,
    DEFAULT_YLIM,
    DEFAULT_CMLIM,
)


def plot_dataset_summary(
    ds: xr.Dataset,
    lat_lon_for_solpos: tuple[float, float] = DEFAULT_LAT_LON_FOR_SOLPOS,
    title: str = DEFAULT_TITLE,
    ylim: tuple[int, int] = DEFAULT_YLIM,
    cmlim: tuple[int, int] = DEFAULT_CMLIM,
    dpi: int = DEFAULT_DPI,
    jpeg_filename: Optional[str] = None,
    show: bool = False,
):  # pylint: disable=R0915  too-many-statements
    """
    Generate a summary plot from the given dataset.
    Code by RYJO, with some typing/formatting/variable naming adjustments.
    :param ds: Dataset to plot.
    :param lat_lon_for_solpos: Lat/Lon for solar position calculation.
    :param title: Title for the plot.
    :param ylim: Limits for the y-axis.
    :param cmlim: Limits passed to pcolormesh.
    :param dpi: DPI to use for the plot.
    :param jpeg_filename: If given, filename to save the plot to.
    :param show: Whether to show the plot.
    """
    # Transpose psd array for plotting
    da = xr.DataArray.transpose(ds.psd)

    # get solar elevation
    # Estimate the solar position with a specific SPA defined with the argument 'method'
    latitude, longitude = lat_lon_for_solpos
    solpos = pvlib.solarposition.get_solarposition(
        ds.time, latitude=latitude, longitude=longitude
    )
    se = solpos.elevation  # isolate solar elevation
    # map elevation to gray scale
    seg = 0 * se  # 0 covers nighttime (black)
    # day (white)
    d = np.squeeze(np.where(se > 0))
    seg.iloc[d] = 1
    # dusk / dawn (gray range)
    d = np.squeeze(np.where(np.logical_and(se <= 0, se >= -12)))
    seg.iloc[d] = 1 - abs(se.iloc[d] / max(abs(se.iloc[d])))
    # Get the indices of the min and max
    seg1 = pd.Series.to_numpy(solpos.elevation)
    minidx = np.squeeze(np.where(seg1 == min(seg1)))
    maxidx = np.squeeze(np.where(seg1 == max(seg1)))

    seg3 = np.tile(seg, (50, 1))

    # plotting variables

    psdlabl = (
        r"Spectrum level (dB re 1 $\mu$Pa$\mathregular{^{2}}$ Hz$\mathregular{^{-1}}$)"
    )
    freqlabl = "Frequency (Hz)"

    # define percentiles
    pctlev = np.array([1, 10, 25, 50, 75, 90, 99])
    # initialize output array
    pctls = np.empty((pctlev.size, ds.frequency.size))
    # get percentiles
    np.nanpercentile(ds.psd, pctlev, axis=0, out=pctls)

    # create a figure
    fig = plt.figure()
    fig.set_figheight(6)
    fig.set_figwidth(12)
    spec = gridspec.GridSpec(
        ncols=2,
        nrows=2,
        width_ratios=[2.5, 1],
        wspace=0.02,
        height_ratios=[0.045, 0.95],
        hspace=0.09,
    )

    # Use more of the available plotting space
    plt.subplots_adjust(left=0.06, right=0.94, bottom=0.12, top=0.89)

    # Spectrogram
    ax0 = fig.add_subplot(spec[2])
    vmin, vmax = cmlim
    sg = plt.pcolormesh(
        ds.time, ds.frequency, da, shading="nearest", cmap="rainbow", vmin=vmin, vmax=vmax
    )
    plt.yscale("log")
    plt.ylim(list(ylim))
    plt.ylabel(freqlabl)
    xl = ax0.get_xlim()
    ax0.set_xticks([])
    # plt.colorbar(location='left', shrink = 0.25, fraction = 0.05)

    # Percentile
    pplabels = ["L99", "L90", "L75", "L50", "L25", "L10", "L1"]
    ax1 = fig.add_subplot(spec[3])
    ax1.yaxis.tick_right()
    ax1.yaxis.set_label_position("right")
    plt.plot(pctls.T, ds.frequency, linewidth=1)
    plt.yscale("log")
    plt.ylim(list(ylim))
    plt.xlabel(psdlabl)
    plt.ylabel(freqlabl)
    plt.legend(loc="lower left", labels=pplabels)

    # day night
    ax3 = fig.add_subplot(spec[0])
    ax3.pcolormesh(seg3, shading="flat", cmap="gray")
    ax3.annotate("Day", (maxidx, 25), weight="bold", ha="center", va="center")
    ax3.annotate(
        "Night", (minidx, 25), weight="bold", color="white", ha="center", va="center"
    )
    ax3.set_xticks([])
    ax3.set_yticks([])

    # colorbar for spectrogram
    r = np.concatenate(np.squeeze(ax0.get_position()))
    cb_ax = fig.add_axes([r[0] + 0.09, r[1] - 0.025, r[2] - 0.25, 0.015])
    q = fig.colorbar(sg, orientation="horizontal", cax=cb_ax)
    q.set_label(psdlabl)

    # time axes for the day/night panel
    # create a dummy time / zero range variable
    timax = fig.add_axes(ax3.get_position(), frameon=False)
    timax.plot(solpos.elevation * 0, "k")
    timax.tick_params(top=True, labeltop=True, bottom=False, labelbottom=False)
    timax.set_ylim(0, 100)
    timax.set_yticks([])
    timax.set_xlim(xl)
    timax.xaxis.set_major_formatter(
        md.ConciseDateFormatter(timax.xaxis.get_major_locator())
    )

    plt.gcf().text(0.5, 0.955, title, fontsize=14, horizontalalignment="center")
    plt.gcf().text(0.65, 0.91, "UTC")

    if jpeg_filename is not None:
        plt.savefig(jpeg_filename, dpi=dpi)
    if show:
        plt.show()
    plt.close(fig)



import os
import pathlib
from dataclasses import dataclass
from math import ceil, floor
from typing import Dict, List, Optional, Tuple
from urllib.parse import ParseResult, urlparse

# import loguru
import numpy as np
import soundfile as sf

from botocore.client import BaseClient, ClientError
from google.cloud.exceptions import NotFound as GsNotFound
from google.cloud.storage import Client as GsClient

from pbp.json_support import get_intersecting_entries, JEntry, parse_json_contents
from pbp.misc_helper import brief_list, map_prefix


@dataclass
class AudioInfo:
    samplerate: int
    channels: int
    subtype: str


class SoundStatus:
    # TODO cleanup!  there's some repetition here wrt FileHelper!

    def __init__(
        self,
        log,  # : loguru.Logger,
        uri: str,
        audio_base_dir: Optional[str],
        audio_path_map_prefix: str,
        audio_path_prefix: str,
        download_dir: Optional[str],
        assume_downloaded_files: bool,
        print_downloading_lines: bool,
        s3_client: Optional[BaseClient] = None,
        gs_client: Optional[GsClient] = None,
    ):
        self.log = log
        self.uri = map_prefix(audio_path_map_prefix, uri)
        self.parsed_uri = urlparse(self.uri)

        self.audio_base_dir = audio_base_dir
        self.audio_path_prefix = audio_path_prefix
        self.s3_client = s3_client
        self.gs_client = gs_client
        self.download_dir: str = download_dir if download_dir else "."
        self.assume_downloaded_files = assume_downloaded_files
        self.print_downloading_lines = print_downloading_lines

        self.error = None

        self.sound_filename = self._get_sound_filename()
        if self.sound_filename is None:
            self.error = "error getting sound filename"
            return

        ai = self._get_audio_info(self.sound_filename)
        if ai is None:
            self.error = "error getting audio info"
            return

        self.audio_info: AudioInfo = ai
        self.sound_file = sf.SoundFile(self.sound_filename)
        self.sound_file_open = True
        self.age = 0  # see _get_sound_status.

    def _get_audio_info(self, sound_filename: str) -> Optional[AudioInfo]:
        try:
            sfi = sf.info(sound_filename)
            return AudioInfo(sfi.samplerate, sfi.channels, sfi.subtype)
        except sf.LibsndfileError as e:
            self.log.error(f"{e}")
            return None

    def _get_sound_filename(self) -> Optional[str]:
        self.log.debug(f"_get_sound_filename: {self.uri=}")
        if self.parsed_uri.scheme in ("s3", "gs"):
            return _download(
                log=self.log,
                parsed_uri=self.parsed_uri,
                download_dir=self.download_dir,
                assume_downloaded_files=self.assume_downloaded_files,
                print_downloading_lines=self.print_downloading_lines,
                s3_client=self.s3_client,
                gs_client=self.gs_client,
            )

        # otherwise assuming local file, so we only inspect the `path` attribute:
        path = self.parsed_uri.path
        if path.startswith("/"):
            sound_filename = f"{self.audio_path_prefix}{path}"
        elif self.audio_base_dir is not None:
            sound_filename = f"{self.audio_base_dir}/{path}"
        else:
            sound_filename = path
        return sound_filename

    def remove_downloaded_file(self):
        if not pathlib.Path(self.sound_filename).exists():
            return

        if (
            self.s3_client is None and self.gs_client is None
        ) or self.parsed_uri.scheme not in ("s3", "gs"):
            self.log.debug(f"No file download involved for {self.uri=}")
            return

        try:
            os.remove(self.sound_filename)
            self.log.debug(f"Removed cached file {self.sound_filename} for {self.uri=}")
        except OSError as e:
            self.log.error(f"Error removing file {self.sound_filename}: {e}")


def _download(
    log,  # : loguru.Logger,
    parsed_uri: ParseResult,
    download_dir: str,
    assume_downloaded_files: bool = False,
    print_downloading_lines: bool = False,
    s3_client: Optional[BaseClient] = None,
    gs_client: Optional[GsClient] = None,
) -> Optional[str]:
    """
    Downloads the given URI to the given download directory.

    NOTE: `assume_downloaded_files` can be set to True to skip downloading files
    that already exist in the download directory.

    One of `s3_client` or `gs_client` must be given.

    :return: Downloaded filename or None if error
    """

    pathlib.Path(download_dir).mkdir(parents=True, exist_ok=True)

    bucket, key, simple = get_bucket_key_simple(parsed_uri)
    local_filename = f"{download_dir}/{simple}"

    if os.path.isfile(local_filename) and assume_downloaded_files:
        log.info(f"ASSUMING ALREADY DOWNLOADED: {bucket=} {key=} to {local_filename}")
        if print_downloading_lines:
            print(f"Assuming already downloaded {parsed_uri.geturl()}")
        return local_filename

    scheme = parsed_uri.scheme
    log.info(f"Downloading {scheme=} {bucket=} {key=} to {local_filename}")
    if print_downloading_lines:
        print(f"downloading {parsed_uri.geturl()}")

    if scheme == "s3":
        assert s3_client is not None
        try:
            s3_client.download_file(bucket, key, local_filename)
            return local_filename
        except ClientError as e:
            log.error(f"Error downloading {scheme=} {bucket}/{key}: {e}")
            return None

    if scheme == "gs":
        assert gs_client is not None
        gs_bucket = gs_client.bucket(bucket)
        blob = gs_bucket.blob(key)
        try:
            blob.download_to_filename(local_filename)
            return local_filename
        except GsNotFound as e:
            log.error(f"Error downloading {scheme=} {bucket}/{key}: {e}")
            return None

    return None


class FileHelper:
    """
    Helps loading audio segments.
    """

    def __init__(
        self,
        log,  # : loguru.Logger,
        json_base_dir: str,
        audio_base_dir: Optional[str] = None,
        audio_path_map_prefix: str = "",
        audio_path_prefix: str = "",
        segment_size_in_mins: int = 1,
        s3_client: Optional[BaseClient] = None,
        gs_client: Optional[GsClient] = None,
        download_dir: Optional[str] = None,
        assume_downloaded_files: bool = False,
        retain_downloaded_files: bool = False,
        print_downloading_lines: bool = False,
    ):
        """

        :param log:
          Logger
        :param audio_base_dir:
          If given, it will be used as base directory for any relative (not starting with a slash)
          `path` attribute in the json entries.
        :param audio_path_map_prefix:
          Prefix mapping to get actual audio uri to be used.
          Example: `s3://pacific-sound-256khz-2022~file:///PAM_Archive/2022`
        :param audio_path_prefix:
          Ad hoc path prefix for sound file locations, e.g. "/Volumes"
        :param segment_size_in_mins:
            The size of each audio segment to extract, in minutes. By default, 1.
        :param s3_client:
            If given, it will be used to handle `s3:` based uris.
        :param gs_client:
            If given, it will be used to handle `gs:` based uris.
        :param download_dir:
            Save downloaded S3 files here if given, otherwise, save in current directory.
        :param assume_downloaded_files:
            If True, skip downloading files that already exist in the download directory.
        :param retain_downloaded_files:
            If True, do not remove downloaded files after use.
        :param print_downloading_lines:
            If True, print "downloading <uri>" lines to console.
        """
        self.log = log

        self.log.info(
            "Creating FileHelper:"
            + f"\n    json_base_dir:           {json_base_dir}"
            + (
                f"\n    audio_base_dir:          {audio_base_dir}"
                if audio_base_dir
                else ""
            )
            + (
                f"\n    audio_path_map_prefix:   '{audio_path_map_prefix}'"
                if audio_path_map_prefix
                else ""
            )
            + (
                f"\n    audio_path_prefix:       '{audio_path_prefix}'"
                if audio_path_prefix
                else ""
            )
            + f"\n    segment_size_in_mins:    {segment_size_in_mins}"
            + f"\n    s3_client:               {'(given)' if s3_client else 'None'}"
            + f"\n    gs_client:               {'(given)' if gs_client else 'None'}"
            + f"\n    download_dir:            {download_dir}"
            + f"\n    assume_downloaded_files: {assume_downloaded_files}"
            + f"\n    retain_downloaded_files: {retain_downloaded_files}"
            + f"\n    print_downloading_lines: {print_downloading_lines}"
            + "\n"
        )
        self.json_base_dir = json_base_dir
        self.audio_base_dir = audio_base_dir
        self.audio_path_map_prefix = audio_path_map_prefix
        self.audio_path_prefix = audio_path_prefix
        self.segment_size_in_mins = segment_size_in_mins
        self.s3_client = s3_client
        self.gs_client = gs_client
        self.download_dir: str = download_dir if download_dir else "."
        self.assume_downloaded_files = assume_downloaded_files
        self.retain_downloaded_files = retain_downloaded_files
        self.print_downloading_lines = print_downloading_lines

        self.sound_cache: Dict[str, SoundStatus] = {}

        # the following set by select_day:
        self.year: Optional[int] = None
        self.month: Optional[int] = None
        self.day: Optional[int] = None
        self.json_entries: Optional[List[JEntry]] = None

    def select_day(self, year: int, month: int, day: int) -> bool:
        """
        Selects the given day for subsequent processing of relevant audio segments.

        :param year:
        :param month:
        :param day:
        :return:  True only if selection was successful
        """

        self.log.info(f"Selecting day: {year:04}{month:02}{day:02}")

        json_uri = f"{self.json_base_dir}/{year:04}/{year:04}{month:02}{day:02}.json"
        json_contents = self._get_json(json_uri)
        if json_contents is None:
            self.log.error(f"{json_uri}: file not found\n")
            return False

        self.year = year
        self.month = month
        self.day = day
        self.json_entries = list(parse_json_contents(json_contents))
        return True

    def get_local_filename(self, uri: Optional[str]) -> Optional[str]:
        """
        Returns the local filename for the given URI, which will be that of
        the downloaded file when the given uri is cloud based.
        """
        if uri is None:
            return None

        parsed_uri = urlparse(uri)
        if parsed_uri.scheme in ("s3", "gs"):
            return _download(
                log=self.log,
                parsed_uri=parsed_uri,
                download_dir=self.download_dir,
                assume_downloaded_files=self.assume_downloaded_files,
                print_downloading_lines=self.print_downloading_lines,
                s3_client=self.s3_client,
                gs_client=self.gs_client,
            )

        return parsed_uri.path

    def day_completed(self):
        """
        Since a process is launched only for a day, we simply clear the cache.
        """
        # first, close all sound files still open:
        num_still_open = 0
        for c_uri, c_ss in list(self.sound_cache.items()):
            # due to some weird issues (when running under dask), let's be extra careful:
            # TODO clean up this!
            if (
                hasattr(c_ss, "sound_file")
                and hasattr(c_ss, "sound_file_open")
                and c_ss.sound_file_open
            ):
                c_ss.sound_file_open = False
                num_still_open += 1
                self.log.debug(f"Closing sound file for cached {c_uri=} {c_ss.age=}")
                c_ss.sound_file.close()
            self.log.debug(
                f"day_completed: closed {num_still_open} sound files that were still open."
            )

        # remove any downloaded files (cloud case):
        if not self.retain_downloaded_files:
            for c_ss in self.sound_cache.values():
                c_ss.remove_downloaded_file()

        self.sound_cache = {}

    def _get_json(self, uri: str) -> Optional[str]:
        parsed_uri = urlparse(uri)
        if parsed_uri.scheme == "s3":
            return self._get_json_s3(parsed_uri)
        #  simply assume local file:
        return self._get_json_local(parsed_uri.path)

    def _get_json_s3(self, parsed_uri: ParseResult) -> Optional[str]:
        local_filename = _download(
            log=self.log,
            parsed_uri=parsed_uri,
            download_dir=self.download_dir,
            assume_downloaded_files=self.assume_downloaded_files,
            print_downloading_lines=self.print_downloading_lines,
            s3_client=self.s3_client,
            gs_client=self.gs_client,
        )
        if local_filename is None:
            return None
        return self._get_json_local(local_filename)

    def extract_audio_segment(
        self, at_hour: int, at_minute: int
    ) -> Optional[Tuple[AudioInfo, np.ndarray]]:
        """
        Extracts the audio segment at the given start time.
        For this it loads and aggregates the relevant audio segments.

        :return:  A tuple (audio_info, audio_segment) or None
        """

        assert self.json_entries is not None
        assert self.year is not None
        assert self.month is not None
        assert self.day is not None

        intersections = get_intersecting_entries(
            self.log,
            self.json_entries,
            self.year,
            self.month,
            self.day,
            at_hour,
            at_minute,
            segment_size_in_mins=self.segment_size_in_mins,
        )

        audio_info: Optional[AudioInfo] = None

        aggregated_segment: Optional[np.ndarray] = None

        prefix = f"({at_hour:02}h:{at_minute:02}m)"
        for intersection in intersections:
            if intersection.duration_secs == 0:
                self.log.warning("No data from intersection")
                continue

            ss = self._get_sound_status(intersection.entry.uri)
            if ss.error is not None:
                return None

            self.log.debug(
                f"    {prefix} {intersection.duration_secs} secs from {ss.sound_filename}"
            )

            if audio_info is not None and not self._check_audio_info(
                audio_info, ss.audio_info
            ):
                return None  # error!

            audio_info = ss.audio_info

            start_sample = floor(intersection.start_secs * audio_info.samplerate)
            num_samples = ceil(intersection.duration_secs * audio_info.samplerate)

            try:
                new_pos = ss.sound_file.seek(start_sample)
                if new_pos != start_sample:
                    # no-data case, let's just read 0 samples to get an empty array:
                    audio_segment = ss.sound_file.read(0)
                else:
                    audio_segment = ss.sound_file.read(num_samples)
                    if len(audio_segment) < num_samples:
                        # partial-data case.
                        self.log.warning(
                            f"!!! partial data: {len(audio_segment)} < {num_samples}"
                        )

            except sf.LibsndfileError as e:
                self.log.error(f"{e}")
                return None

            if aggregated_segment is None:
                aggregated_segment = audio_segment
            else:
                aggregated_segment = np.concatenate((aggregated_segment, audio_segment))

        if aggregated_segment is not None:
            assert audio_info is not None
            return audio_info, aggregated_segment
        return None

    def _check_audio_info(self, ai1: AudioInfo, ai2: AudioInfo) -> bool:
        if ai1.samplerate != ai2.samplerate:
            self.log.error(
                f"UNEXPECTED: sample rate mismatch: {ai1.samplerate} vs {ai2.samplerate}"
            )
            return False
        if ai1.channels != ai2.channels:
            self.log.error(
                f"UNEXPECTED: channel count mismatch: {ai1.channels} vs {ai2.channels}"
            )
            return False
        if ai1.subtype != ai2.subtype:
            self.log.error(
                f"UNEXPECTED: subtype mismatch: {ai1.subtype} vs {ai2.subtype}"
            )
            return False
        return True

    def _get_sound_status(self, uri: str) -> SoundStatus:
        """
        Returns a SoundStatus object for the given uri.
        Internally, the 'age' attribute helps to keep the relevant files open
        as long as recently used. Note that traversal of the files indicated in the
        JSON array happens in a monotonically increasing order in time, so we
        can increment the 'age' for all entries in the cache except for the uri
        just requested.

        :param uri:
        :return:
        """
        self.log.debug(f"_get_sound_status: {uri=}")
        ss = self.sound_cache.get(uri)
        if ss is None:
            # currently cached ones get a bit older:
            for c_ss in self.sound_cache.values():
                c_ss.age += 1

            self.log.debug(f"SoundStatus: creating for {uri=}")
            ss = SoundStatus(
                log=self.log,
                uri=uri,
                audio_base_dir=self.audio_base_dir,
                audio_path_map_prefix=self.audio_path_map_prefix,
                audio_path_prefix=self.audio_path_prefix,
                download_dir=self.download_dir,
                assume_downloaded_files=self.assume_downloaded_files,
                print_downloading_lines=self.print_downloading_lines,
                s3_client=self.s3_client,
                gs_client=self.gs_client,
            )
            self.sound_cache[uri] = ss
        else:
            self.log.debug(f"SoundStatus: already available for {uri=}")

        # close and remove files in the cache that are not fresh enough in terms
        # of not being recently used
        for c_uri, c_ss in list(self.sound_cache.items()):
            if uri != c_uri and c_ss.age > 2 and c_ss.sound_file_open:
                self.log.debug(
                    f"Closing sound file for cached uri={c_uri} age={c_ss.age}"
                )
                c_ss.sound_file.close()
                c_ss.sound_file_open = False
                if not self.retain_downloaded_files:
                    c_ss.remove_downloaded_file()

        def log_msg():
            c_sss = self.sound_cache.values()
            open_files = len([c_ss for c_ss in c_sss if c_ss.sound_file_open])
            ages = [c_ss.age for c_ss in c_sss]
            return f"{open_files=}  ages={brief_list(ages)}"

        self.log.opt(lazy=True).debug("_get_sound_status: {}", log_msg)

        return ss

    def _get_json_local(self, filename: str) -> Optional[str]:
        try:
            with open(filename, "r", encoding="UTF-8") as f:
                return f.read()
        except IOError as e:
            self.log.error(f"Error reading {filename}: {e}")
            return None


def get_bucket_key_simple(parsed_uri: ParseResult) -> Tuple[str, str, str]:
    bucket = parsed_uri.netloc
    key = parsed_uri.path.lstrip("/")
    simple = key.split("/")[-1] if "/" in key else key
    assert "/" not in simple, f"Unexpected simple_filename: '{simple}'"
    return bucket, key, simple



import json
from collections import OrderedDict
from typing import Any, Dict, Optional

import xarray as xr
import yaml


def parse_attributes(contents: str, suffix: str) -> OrderedDict[str, Any]:
    """
    Parses given contents into a dictionary of attributes.
    :param contents:
        String with valid JSON or YAML contents.
    :param suffix:
        Used to determine the content format.
    :return:
        The parsed attributes.
    """
    if suffix == ".json":
        return json.loads(contents, object_pairs_hook=OrderedDict)
    if suffix in (".yaml", ".yml"):
        return yaml.load(contents, Loader=yaml.SafeLoader)
    raise ValueError(f"Unrecognized contents for format: {suffix}")


class MetadataHelper:
    def __init__(
        self,
        log,  # : loguru.Logger,
        global_attributes: Optional[OrderedDict[str, Any]] = None,
        variable_attributes: Optional[OrderedDict[str, Any]] = None,
    ):
        self.log = log
        self._global_attrs: OrderedDict[str, Any] = global_attributes or OrderedDict()
        self._var_attrs: OrderedDict[str, Any] = variable_attributes or OrderedDict()

    def set_some_global_attributes(self, attrs: Dict[str, Any]):
        for k, v in attrs.items():
            self._global_attrs[k] = v

    def get_global_attributes(self) -> OrderedDict[str, Any]:
        return self._global_attrs

    def add_variable_attributes(self, da: xr.DataArray, var_attribute_name: str):
        if var_attribute_name in self._var_attrs:
            keys = []
            for k, v in self._var_attrs[var_attribute_name].items():
                da.attrs[k] = v
                keys.append(k)
            self.log.debug(
                f"For variable '{var_attribute_name}', added attributes: {keys}"
            )
        else:
            self.log.error(f"Unrecognized {var_attribute_name=}")


def replace_snippets(
    attributes: OrderedDict[str, Any], snippets: Dict[str, str]
) -> OrderedDict[str, Any]:
    """
    Replaces snippets in any entries with values of type string.
    :param attributes:
        Attribute dictionary to replace snippets in.
    :param snippets:
        Example: { "{{PyPAM_version}}": "0.2.0" }
    :return:
        A new dictionary with the snippets replaced.
    """
    result = OrderedDict()
    for k, v in attributes.items():
        if isinstance(v, str):
            for snippet, replacement in snippets.items():
                v = v.replace(snippet, replacement)
        result[k] = v
    return result



from argparse import ArgumentParser, RawTextHelpFormatter


from pbp import get_pbp_version


def parse_arguments():
    description = "Process ocean audio data archives to daily analysis products of hybrid millidecade spectra using PyPAM."
    example = """
Examples:
    pbp --json-base-dir=tests/json \\
        --audio-base-dir=tests/wav \\
        --date=20220902 \\
        --output-dir=output
    """

    parser = ArgumentParser(
        description=description, epilog=example, formatter_class=RawTextHelpFormatter
    )

    parser.add_argument(
        "--version",
        action="version",
        version=get_pbp_version(),
    )

    parser.add_argument(
        "--json-base-dir",
        type=str,
        metavar="dir",
        required=True,
        help="JSON base directory",
    )

    parser.add_argument(
        "--audio-base-dir",
        type=str,
        metavar="dir",
        default=None,
        help="Audio base directory. By default, none",
    )

    parser.add_argument(
        "--global-attrs",
        type=str,
        metavar="uri",
        default=None,
        help="URI of JSON file with global attributes to be added to the NetCDF file.",
    )

    parser.add_argument(
        "--set-global-attr",
        type=str,
        nargs=2,
        default=None,
        metavar=("key", "value"),
        dest="set_global_attrs",
        action="append",
        help="Replace {{key}} with the given value for every occurrence of {{key}}"
        " in the global attrs file.",
    )

    parser.add_argument(
        "--variable-attrs",
        type=str,
        metavar="uri",
        default=None,
        help="URI of JSON file with attributes to associate to the variables in the NetCDF file.",
    )

    parser.add_argument(
        "--audio-path-map-prefix",
        type=str,
        metavar="from~to",
        default="",
        help="Prefix mapping to get actual audio uri to be used."
        " Example: 's3://pacific-sound-256khz-2022~file:///PAM_Archive/2022'.",
    )

    parser.add_argument(
        "--audio-path-prefix",
        type=str,
        metavar="dir",
        default="",
        help="Ad hoc path prefix for sound file location, for example, /Volumes."
        " By default, no prefix applied.",
    )

    parser.add_argument(
        "--date",
        type=str,
        required=True,
        metavar="YYYYMMDD",
        help="The date to be processed.",
    )

    parser.add_argument(
        "--voltage-multiplier",
        type=float,
        default=None,
        metavar="value",
        help="Applied on the loaded signal.",
    )

    parser.add_argument(
        "--sensitivity-uri",
        type=str,
        default=None,
        metavar="file",
        help="URI of sensitivity NetCDF for calibration of result. "
        "Has precedence over --sensitivity-flat-value.",
    )

    parser.add_argument(
        "--sensitivity-flat-value",
        type=float,
        default=None,
        metavar="value",
        help="Flat sensitivity value to be used for calibration.",
    )

    parser.add_argument(
        "--output-dir",
        type=str,
        metavar="dir",
        required=True,
        help="Output directory",
    )

    parser.add_argument(
        "--output-prefix",
        type=str,
        metavar="prefix",
        default="milli_psd_",
        help="Output filename prefix",
    )

    parser.add_argument(
        "--s3",
        default=False,
        action="store_true",
        help="s3 access involved.",
    )

    parser.add_argument(
        "--gs",
        default=False,
        action="store_true",
        help="gs access involved.",
    )

    parser.add_argument(
        "--download-dir",
        type=str,
        metavar="dir",
        default=None,
        help="Directory for any downloads (e.g., when s3 or gs is involved).",
    )

    parser.add_argument(
        "--assume-downloaded-files",
        default=False,
        action="store_true",
        help="If any destination file for a download exists, assume it was downloaded already.",
    )

    parser.add_argument(
        "--retain-downloaded-files",
        default=False,
        action="store_true",
        help="Do not remove any downloaded files after use.",
    )

    parser.add_argument(
        "--max-segments",
        type=int,
        default=0,
        metavar="num",
        help="Test convenience: limit number of segments to process. By default, 0 (no limit).",
    )

    parser.add_argument(
        "--subset-to",
        type=int,
        nargs=2,
        default=None,
        metavar=("lower", "upper"),
        help="Subset the resulting PSD to [lower, upper), in terms of central frequency.",
    )

    return parser.parse_args()



from dataclasses import dataclass
from datetime import datetime
from typing import cast, List, Optional, Tuple

import numpy as np
import pypam.signal as sig

import xarray as xr
from pypam import utils

from pbp.misc_helper import brief_list


@dataclass
class ProcessResult:
    """
    The result of processing a day of audio segments when at least one actual segment
    has been captured.
    """

    psd_da: xr.DataArray
    effort_da: xr.DataArray


@dataclass
class _CapturedSegment:
    dt: datetime
    num_secs: float
    spectrum: Optional[np.ndarray]


class PypamSupport:
    def __init__(
        self,
        log,  #: loguru.Logger
    ) -> None:
        """
        Creates a helper to process audio segments for a given day,
        resulting in the aggregated hybrid millidecade PSD product.

        The created instance should be used only for a day of processing.

        The overall call sequence is:

        `add_missing_segment` can be called before calling `set_parameters`.
        This allows capturing initial segments with no data for the day.

        Call `set_parameters` as soon as you get the first audio segment for a day.
        Such segment is used to determine the sampling frequency.

        Then you can call `add_segment` or `add_missing_segment` as appropriate
        for each subsequent segment until covering the day.

        When all segments have been captured, call `process_captured_segments`
        to get the result.
        """
        self.log = log

        # to capture reported segments (missing and otherwise):
        self._captured_segments: List[_CapturedSegment] = []
        self._num_actual_segments: int = 0

        # Determined from any actual segment:
        self._fbands: Optional[np.ndarray] = None

        # The following determined when `set_parameters` is called:

        self.fs: Optional[int] = None
        self._nfft: Optional[int] = None
        self._subset_to: Optional[Tuple[int, int]] = None
        self._bands_limits: List[float] = []
        self._bands_c: List[float] = []

    def set_parameters(
        self, fs: int, nfft: int = 0, subset_to: Optional[Tuple[int, int]] = None
    ):
        """
        Call this as soon as you get the first audio segment for a day,
        in particular, to set the sampling frequency used for subsequent calculations.

        :param fs:
            Sampling frequency.
        :param nfft:
            Number of samples to use for the FFT. If 0, it will be set to `fs`.
        :param subset_to:
            If not None, the product for a day will get the resulting PSD
            subset to `[lower, upper)`, in terms of central frequency.
        """
        assert fs > 0
        self.fs = fs
        self._nfft = nfft if nfft > 0 else self.fs

        self._subset_to = subset_to
        band = [0, self.fs / 2]  # for now.

        self.log.debug(f"PypamSupport: {subset_to=} {band=}")

        self._bands_limits, self._bands_c = utils.get_hybrid_millidecade_limits(
            band=band, nfft=self._nfft
        )

    @property
    def parameters_set(self) -> bool:
        """
        True if `set_parameters` has already been called.
        """
        return self.fs is not None

    def add_missing_segment(self, dt: datetime):
        """
        Adds a missing segment to the ongoing aggregation.
        This method can be called even before `set_parameters` is called.

        :param dt:
            The datetime of the start of the missing segment.
        """
        self._captured_segments.append(_CapturedSegment(dt, 0, None))
        self.log.debug(f"  captured segment: {dt}  (NO DATA)")

    def add_segment(self, dt: datetime, data: np.ndarray):
        """
        Adds an audio segment to the ongoing aggregation.

        `set_parameters` must have been called first.

        :param dt:
            The datetime of the start of the segment.
        :param data:
            The audio data.
        """
        assert self.parameters_set
        assert self.fs is not None
        assert self._nfft is not None

        self._fbands, spectrum = _get_spectrum(data, self.fs, self._nfft)
        num_secs = len(data) / self.fs
        self._captured_segments.append(_CapturedSegment(dt, num_secs, spectrum))
        self._num_actual_segments += 1
        self.log.debug(f"  captured segment: {dt}")

    def process_captured_segments(
        self,
        sensitivity_da: Optional[xr.DataArray] = None,
    ) -> Optional[ProcessResult]:
        """
        Gets the resulting hybrid millidecade bands for all captured segments.
        At least one actual segment must have been captured, otherwise None is returned.

        :param sensitivity_da:
            If given, it will be used to calibrate the result.
        :return:
            Result if at least an actual segment was captured, None otherwise.
        """
        if self._num_actual_segments == 0:
            return None

        # Use any actual segment to determine NaN spectrum for the missing segments:
        actual = next(s for s in self._captured_segments if s.spectrum is not None)
        assert actual is not None, "unexpected: no actual segment found"
        assert actual.spectrum is not None, "unexpected: no actual.spectrum"
        nan_spectrum = np.full(len(actual.spectrum), np.nan)

        # gather resulting variables:
        times: List[np.int64] = []
        effort: List[np.float32] = []
        spectra: List[np.ndarray] = []
        for cs in self._captured_segments:
            times.append(np.int64(cs.dt.timestamp()))
            effort.append(np.float32(cs.num_secs))

            spectrum = nan_spectrum if cs.spectrum is None else cs.spectrum
            self.log.debug(f"  spectrum for: {cs.dt} (effort={cs.num_secs})")
            spectra.append(spectrum)

        self.log.info("Aggregating results ...")
        psd_da = self._get_aggregated_milli_psd(
            times=times,
            spectra=spectra,
            sensitivity_da=sensitivity_da,
        )

        effort_da = xr.DataArray(
            data=effort,
            dims=["time"],
            coords={"time": psd_da.time},
        )
        return ProcessResult(psd_da, effort_da)

    def _get_aggregated_milli_psd(
        self,
        times: List[np.int64],
        spectra: List[np.ndarray],
        sensitivity_da: Optional[xr.DataArray] = None,
    ) -> xr.DataArray:
        # Convert the spectra to a DataArray
        psd_da = xr.DataArray(
            data=spectra,
            coords={"time": times, "frequency": self._fbands},
            dims=["time", "frequency"],
        )

        psd_da = self._spectra_to_bands(psd_da)
        self.log.debug(f"  {psd_da.frequency_bins=}")
        psd_da = self._apply_sensitivity_if_given(psd_da, sensitivity_da)

        # just need single precision:
        psd_da = psd_da.astype(np.float32)
        psd_da["frequency"] = psd_da.frequency_bins.astype(np.float32)
        del psd_da["frequency_bins"]

        milli_psd = psd_da
        milli_psd.name = "psd"

        self.log.info(f"Resulting milli_psd={milli_psd}")

        return milli_psd

    def _apply_sensitivity_if_given(
        self,
        psd_da: xr.DataArray,
        sensitivity_da: Optional[xr.DataArray],
    ) -> xr.DataArray:
        psd_da = cast(xr.DataArray, 10 * np.log10(psd_da))

        # NOTE: per slack discussion today 2023-05-23,
        # apply _addition_ of the given sensitivity
        # (previously, subtraction)
        # 2023-06-12: Back to subtraction (as we're focusing on MARS data at the moment)
        # TODO but this is still one pending aspect to finalize.
        # 2023-08-03: sensitivity_flat_value is handled upstream now.

        if sensitivity_da is not None:
            freq_subset = sensitivity_da.interp(frequency=psd_da.frequency_bins)
            self.log.info(
                f"  Applying sensitivity({len(freq_subset.values)})={freq_subset}"
            )
            psd_da -= freq_subset.values

        return psd_da

    def _spectra_to_bands(self, psd_da: xr.DataArray) -> xr.DataArray:
        assert self.fs is not None
        assert self._nfft is not None

        bands_limits, bands_c = self._bands_limits, self._bands_c
        if self._subset_to is not None:
            bands_limits, bands_c = self._adjust_limits(
                bands_limits, bands_c, self._subset_to
            )

        def print_array(name: str, arr: List[float]):
            self.log.info(f"{name} ({len(arr)}) = {brief_list(arr)}")

        print_array("       bands_c", bands_c)
        print_array("  bands_limits", bands_limits)

        psd_da = utils.spectra_ds_to_bands(
            psd_da,
            bands_limits,
            bands_c,
            fft_bin_width=self.fs / self._nfft,
            db=False,
        )

        psd_da = psd_da.drop_vars(["lower_frequency", "upper_frequency"])
        return psd_da

    def _adjust_limits(
        self, bands_limits: List[float], bands_c: List[float], subset_to: Tuple[int, int]
    ) -> Tuple[List[float], List[float]]:
        start_hz, end_hz = subset_to
        self.log.info(f"Subsetting to [{start_hz:,}, {end_hz:,})Hz")

        start_index = 0
        while start_index < len(bands_c) and bands_c[start_index] < start_hz:
            start_index += 1
        end_index = start_index
        while end_index < len(bands_c) and bands_c[end_index] < end_hz:
            end_index += 1
        bands_c = bands_c[start_index:end_index]
        new_bands_c_len = len(bands_c)
        bands_limits = bands_limits[start_index : start_index + new_bands_c_len + 1]

        return bands_limits, bands_c


def _get_spectrum(data: np.ndarray, fs: int, nfft: int) -> Tuple[np.ndarray, np.ndarray]:
    signal = sig.Signal(data, fs=fs)
    signal.set_band(None)
    fbands, spectrum, _ = signal.spectrum(
        scaling="density", nfft=nfft, db=False, overlap=0.5, force_calc=True
    )
    return fbands, spectrum



import copy
import sys
import loguru
import os
import pathlib

from typing import Optional, Tuple


def create_logger(
    log_filename_and_level: Optional[Tuple[str, str]] = None,
    console_level: Optional[str] = None,
):  # -> loguru.Logger:  # poetry complains about not recognizing this type :(
    """
    Create logger instance, typically to be used for a day of processing.

    NOTE: The created object should be passed around to all relevant functions.
    Do not use `from loguru import logger` (or similar) to obtain the logger.

    :param log_filename_and_level:
        (filename, level) tuple or None to disable file logging.
    :param console_level:
        The log level for the console, or None to disable console logging.
    """

    fmt = "{time} {level} {message}"
    loguru.logger.remove()
    log = copy.deepcopy(loguru.logger)
    if log_filename_and_level is not None:
        log_filename, log_level = log_filename_and_level
        # create log_filename's parent directory if needed:
        parent_dir = pathlib.Path(log_filename).parent
        pathlib.Path(parent_dir).mkdir(parents=True, exist_ok=True)
        file_fmt = fmt
        if os.getenv("EXCLUDE_LOG_TIME", "no") == "yes":
            # test convenience to facilitate local diffing of log files
            file_fmt = "{level} {message}"
        log.add(
            sink=open(log_filename, "w"), level=log_level, format=file_fmt, enqueue=True
        )

    if console_level is not None:
        log.add(
            sink=sys.stderr, level=console_level, format=fmt, colorize=True, enqueue=True
        )

    return log
